#!/usr/bin/python
#
# Copyright (c) 2016 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import optparse

import pysam

import paleomix
import paleomix.ui

from paleomix.config import \
    ConfigError, \
    PerHostValue, \
    PerHostConfig, \
    migrate_config

import paleomix.common.fileutils as fileutils
import paleomix.tools.zonkey.database as database


_USAGE = """USAGE:
{0} run <SampleDB> <samples.txt> [<destination>]
{0} run <SampleDB> <nuclear.bam> [<destination>]
{0} run <SampleDB> <mitochondrial.bam> [<destination>]
{0} run <SampleDB> <nuclear.bam> <mitochondrial.bam> <destination>
{0} dryrun <SampleDB> [...]
{0} mito <SampleDB> <destination>
"""

# List of valid commands / aliases, currently mostly undocumented
_CMD_ALIASES = {
    "mito": "mito",
    "mitochondria": "mito",
    "mitochondrial": "mito",
    "nuc": "run",
    "nuclear": "run",
    "run": "run",
    "dryrun": "dryrun",
    "dry_run": "dryrun",
}


def print_usage(out=sys.stderr):
    out.write(_USAGE.format("paleomix zonkey"))


def parse_config(argv):
    migrate_config()

    config, args = _parse_arguments(argv)
    if not args:
        print_usage()
        return

    config.command = _CMD_ALIASES.get(args[0])
    if config.command is None:
        sys.stderr.write("ERROR: Unknown command %r\n" % (args[0],))
        return
    elif config.command == "dryrun":
        config.command = "run"
        config.dry_run = True

    return parse_run_config(config, args[1:])


def parse_run_config(config, args):
    if not (2 <= len(args) <= 4):
        print_usage()
        return

    config.multisample = False
    config.tablefile = args[0]

    try:
        config.database = database.ZonkeyDB(config.tablefile)
    except database.ZonkeyDBError, error:
        sys.stderr.write("ERROR reading database %r: %s\n"
                         % (config.tablefile, error))
        return

    known_samples = set(config.database.samples) | set(("Sample",))
    unknown_samples = set(config.treemix_outgroup) - known_samples
    if unknown_samples:
        sys.stderr.write("ERROR: Argument --treemix-outgroup includes unknown "
                         "sample(s): %s; known samples are %s. Note that "
                         "names are case-sensitive.\n"
                         % (", ".join(map(repr, sorted(unknown_samples))),
                            ", ".join(map(repr, sorted(known_samples)))))
        return

    if config.command == "mito":
        if len(args) != 2:
            sys.stderr.write("ERROR: Wrong number of arguments!\n")
            print_usage()
            return

        config.destination = args[1]
        config.samples = {}
    elif len(args) == 2:
        filename = args[1]
        config.destination = fileutils.swap_ext(filename, ".zonkey")

        if not os.path.isfile(filename):
            sys.stderr.write("ERROR: Not a valid filename: %r\n" % (filename,))
            return
        elif _is_bamfile(filename):
            # Called as either of
            #   zonkey run <SampleDB> <nuclear.bam>
            #   zonkey run <SampleDB> <mitochondrial.bam>
            config.samples = {"-": {"Root": config.destination,
                                    "Files": [filename]}}
        else:
            config.multisample = True
            if not _read_sample_table(config, filename):
                return
    elif 3 <= len(args) <= 4:
        root = args[-1]
        if os.path.exists(root) and not os.path.isdir(root):
            sys.stderr.write("ERROR: Missing destination folder.\n")
            print_usage()
            return

        config.destination = root

        if len(args) == 3:
            # zonkey run <SampleDB> <nuclear.bam> <destination>
            # zonkey run <SampleDB> <mitochondrial.bam> <destination>
            # zonkey run <SampleDB> <samples.txt> <destination>
            filename = args[-2]

            if not os.path.isfile(filename):
                sys.stderr.write("ERROR: Not a valid filename: %r\n"
                                 % (filename,))
                return
            elif _is_bamfile(filename):
                # Called as either of
                #   zonkey run <SampleDB> <nuclear.bam>
                #   zonkey run <SampleDB> <mitochondrial.bam>
                config.samples = {"-": {"Root": config.destination,
                                        "Files": [filename]}}
            else:
                config.multisample = True
                if not _read_sample_table(config, filename):
                    return
        else:
            # zonkey run <SampleDB> <nuclear.bam> <mitochondrial.bam> <dst>
            config.destination = root
            config.samples = {"-": {"Root": root, "Files": args[1:-1]}}
    else:
        raise RuntimeError("Unhandled number of args in parse_config: %i\n"
                           % (len(args),))

    # Identify (mito or nuc?) and validate BAM files provided by user
    if not _process_samples(config):
        return

    return config


def _is_bamfile(filename):
    """Returns true if a file is a BAM file, false otherwise.
    """
    try:
        with pysam.Samfile(filename, "rb"):
            return True
    except ValueError:
        return False
    except IOError:
        return False


def _process_samples(config):
    for name, info in sorted(config.samples.items()):
        files = {}

        if name == "-":
            sys.stderr.write("Validating unnamed sample ...\n")
        else:
            sys.stderr.write("Validating sample %r ...\n" % (name,))

        for filename in info.pop("Files"):
            filetype = config.database.validate_bam(filename)
            if not filetype:
                sys.stderr.write("ERROR: File is not a valid BAM file: %r\n"
                                 % (filename,))
                return False

            if filetype.is_nuclear and filetype.is_mitochondrial:
                if "Nuc" in files:
                    sys.stderr.write("ERROR: Two nuclear BAMs specified!\n")
                    return False
                elif "Mito" in files:
                    sys.stderr.write("WARNING: Nuclear + mitochondrial BAM, "
                                     "and mitochondrial BAM specified; the "
                                     "mitochondrial genome in the first BAM "
                                     "will not be used!\n")

                files["Nuc"] = filename
                files.setdefault("Mito", filename)
            elif filetype.is_nuclear:
                if "Nuc" in files:
                    sys.stderr.write("ERROR: Two nuclear BAMs specified!\n")
                    return False

                files["Nuc"] = filename
            elif filetype.is_mitochondrial:
                if "Mito" in files:
                    sys.stderr.write("ERROR: Two nuclear BAMs specified!\n")
                    return False

                files["Mito"] = filename
            else:
                sys.stderr.write("ERROR: BAM does not contain known nuclear "
                                 "or mitochondrial contigs: %r\n"
                                 % (filename,))
                return False

#        print_info("Found mt sequence; %r, %i bp, %i bp padding.\n"
#                   % (bam_contig, bam_length, bam_padding))

        config.samples[name]["Files"] = files

    return True


def _read_sample_table(config, filename):
    """Parses a 2 - 3 column tab-seperated table containing, on each row, a
    name to be used for a sample in the first row, and then the paths two
    either one or to two BAM files, which must represent a single nuclear or
    a single mitochondrial alignment (2 columns), or both (3 columns).
    """
    sys.stderr.write("Reading table of samples from %r\n" % (filename,))

    samples = config.samples = {}
    with fileutils.open_ro(filename) as handle:
        for linenum, line in enumerate(handle, start=1):
            if not line.strip() or line.lstrip().startswith("#"):
                continue

            fields = filter(None, line.rstrip('\r\n').split('\t'))
            if len(fields) not in (2, 3):
                sys.stderr.write("Error reading sample table (%r) at line %i; "
                                 "expected 2 or 3 columns, found %i; please "
                                 "correct file before continuing."
                                 % (filename, linenum, len(fields)))
                return

            name = fields[0]
            if name in samples:
                sys.stderr.write("Duplicate sample name found in sample table "
                                 "(%r) at line %i: %r. All sample names must "
                                 "be unique!" % (filename, linenum, name))
                return

            samples[name] = {"Root": os.path.join(config.destination, name),
                             "Files": fields[1:]}

    return True


def _parse_arguments(argv):
    per_host_cfg = PerHostConfig("hybrid_pipeline")

    usage_str = "%prog <command> [options] <SampleDB> <bam/sam> [destination]"
    version_str = "%%prog v%s" % (paleomix.__version__,)
    parser = optparse.OptionParser(usage=usage_str,
                                   version=version_str)

    group = optparse.OptionGroup(parser, "Program options")
    group.add_option("--downsample-to", type=int, default=1000000,
                     help="Number of reads to use for analyses; if 0, no "
                          "downsampling is performed [default: %default]")
    group.add_option("--admixture-replicates", type=int, default=1,
                     help="Number of admixture replicates to run, before "
                          "the result with the highest likelihood [%default]")
    group.add_option("--treemix-k", type=int, default=None,
                     help="Value passed to treemix's -k option; number of "
                          "SNPs per block for estimation of the covariance "
                          "matrix. If not set manually, a value will be "
                          "estimated assuming an even distriution of SNPs "
                          "[%default]")
    group.add_option("--treemix-outgroup", default="",
                     help="Comma-seperated list of samples to use as the "
                          "outgroup when running TreeMix; note that these "
                          "must form a monophyletic clade, or TreeMix will "
                          "not execute.")

    group.add_option("--admixture-only", help=optparse.SUPPRESS_HELP,
                     default=False, action="store_true")
    group.add_option("--indep", nargs=3, help=optparse.SUPPRESS_HELP)
    group.add_option("--indep-pairwise", nargs=3, help=optparse.SUPPRESS_HELP)
    group.add_option("--indep-pairphase", nargs=3, help=optparse.SUPPRESS_HELP)

    parser.add_option_group(group)

    paleomix.ui.add_optiongroup(parser,
                                ui_default=PerHostValue("progress"),
                                color_default=PerHostValue("on"))
    paleomix.logger.add_optiongroup(parser, default=PerHostValue("warning"))

    group = optparse.OptionGroup(parser, "Pipeline")
    group.add_option("--dry-run", action="store_true", default=False,
                     help="If passed, only a dry-run in performed, the "
                          "dependency tree is printed, and no tasks are "
                          "executed.")
    group.add_option("--max-threads",
                     type=int, default=PerHostValue(1),
                     help="Maximum number of threads to use [%default]")
    group.add_option("--list-output-files",
                     action="store_true", default=False,
                     help="List all files to be generated during a run.")
    group.add_option("--list-executables", action="store_true", default=False,
                     help="List all executables required by the pipeline, "
                          "with version requirements (if any).")
    group.add_option("--to-dot-file", dest="dot_file",
                     help="Write dependency tree to the specified dot-file.")
    parser.add_option_group(group)

    config, args = per_host_cfg.parse_args(parser, argv)
    paleomix.ui.set_ui_colors(config.ui_colors)

    indep_opts = (config.indep, config.indep_pairwise, config.indep_pairphase)
    if sum(bool(value) for value in indep_opts) > 1:
        parser.error("Multiple --indep* options specified!")

    if config.indep:
        config.indep_params = config.indep
        config.indep = 'indep'
    elif config.indep_pairwise:
        config.indep_params = config.indep_pairwise
        config.indep = 'indep-pairwise'
    elif config.indep_pairphase:
        config.indep_params = config.indep_pairphase
        config.indep = 'indep-pairphase'

    config.treemix_outgroup \
        = tuple(filter(None, sorted(config.treemix_outgroup.split(","))))

    return config, args



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import getpass

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import \
    AtomicJavaCmdBuilder, \
    create_customizable_cli_parameters, \
    use_customizable_cli_parameters
from paleomix.common.fileutils import \
    swap_ext, \
    try_rmtree, \
    try_remove, \
    reroot_path, \
    describe_files
from paleomix.common.utilities import \
    safe_coerce_to_tuple
import paleomix.common.versions as versions


class PicardNode(CommandNode):
    """Base class for nodes using Picard Tools; adds an additional cleanup
    step, in order to allow the jars to be run using the same temporary folder
    as any other commands associated with the node. This is nessesary as some
    Picard tools create a large number of temporary files, leading to potential
    performance issues if these are located in the same folder.
    """

    def _teardown(self, config, temp):
        # Picard creates a folder named after the user in the temp-root
        try_rmtree(os.path.join(temp, getpass.getuser()))
        # Some JREs may create a folder for temporary performance counters
        try_rmtree(os.path.join(temp, "hsperfdata_" + getpass.getuser()))

        CommandNode._teardown(self, config, temp)


class ValidateBAMNode(PicardNode):
    @create_customizable_cli_parameters
    def customize(cls, config, input_bam, output_log=None, dependencies=()):
        params = picard_command(config, "ValidateSamFile")

        params.set_option("I", "%(IN_BAM)s", sep="=")

        output_log = output_log or swap_ext(input_bam, ".validated")
        params.set_kwargs(IN_BAM=input_bam,
                          OUT_STDOUT=output_log)

        return {"command": params,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        description = "<Validate BAM: '%s'>" % (parameters.input_bam,)
        PicardNode.__init__(self,
                            command=parameters.command.finalize(),
                            description=description,
                            dependencies=parameters.dependencies)


class BuildSequenceDictNode(PicardNode):
    @create_customizable_cli_parameters
    def customize(cls, config, reference, dependencies=()):
        params = picard_command(config, "CreateSequenceDictionary")

        params.set_option("R", "%(TEMP_OUT_REF)s", sep="=")
        params.set_option("O", "%(OUT_DICT)s", sep="=")
        params.set_kwargs(IN_REF=reference,
                          TEMP_OUT_REF=os.path.basename(reference),
                          OUT_DICT=swap_ext(reference, ".dict"))

        return {"command": params,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._in_reference = os.path.abspath(parameters.reference)
        description = "<SequenceDictionary: '%s'>" % (parameters.reference,)

        PicardNode.__init__(self,
                            command=parameters.command.finalize(),
                            description=description,
                            dependencies=parameters.dependencies)

    def _setup(self, _config, temp):
        os.symlink(self._in_reference, reroot_path(temp, self._in_reference))


class MarkDuplicatesNode(PicardNode):
    @create_customizable_cli_parameters
    def customize(cls, config, input_bams, output_bam, output_metrics=None,
                  keep_dupes=False, dependencies=()):
        params = picard_command(config, "MarkDuplicates")

        # Create .bai index, since it is required by a lot of other programs
        params.set_option("CREATE_INDEX", "True", sep="=")

        params.set_option("OUTPUT", "%(OUT_BAM)s", sep="=")
        params.set_option("METRICS_FILE", "%(OUT_METRICS)s", sep="=")
        params.add_multiple_options("I", input_bams, sep="=")

        if not keep_dupes:
            # Remove duplicates from output by default to save disk-space
            params.set_option("REMOVE_DUPLICATES", "True",
                              sep="=", fixed=False)

        output_metrics = output_metrics or swap_ext(output_bam, ".metrics")
        params.set_kwargs(OUT_BAM=output_bam,
                          OUT_BAI=swap_ext(output_bam, ".bai"),
                          OUT_METRICS=output_metrics)

        return {"command": params,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        description = "<MarkDuplicates: %s>" \
            % (describe_files(parameters.input_bams),)
        PicardNode.__init__(self,
                            command=parameters.command.finalize(),
                            description=description,
                            dependencies=parameters.dependencies)


class MergeSamFilesNode(PicardNode):
    @create_customizable_cli_parameters
    def customize(cls, config, input_bams, output_bam, dependencies=()):
        params = picard_command(config, "MergeSamFiles")

        params.set_option("OUTPUT", "%(OUT_BAM)s", sep="=")
        params.set_option("CREATE_INDEX", "True", sep="=")
        params.set_option("SO", "coordinate", sep="=", fixed=False)
        params.add_multiple_options("I", input_bams, sep="=")

        params.set_kwargs(OUT_BAM=output_bam,
                          OUT_BAI=swap_ext(output_bam, ".bai"))

        return {"command": params,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        description = "<Merge BAMs: %i file(s) -> '%s'>" \
            % (len(parameters.input_bams), parameters.output_bam)
        PicardNode.__init__(self,
                            command=parameters.command.finalize(),
                            description=description,
                            dependencies=parameters.dependencies)


class MultiBAMInput(object):
    """Container used to ease processing of 1 or more BAM files; used in
    conjunctin with MultiBAMInputNode.
    """

    def __init__(self, config, input_bams, pipename="input.bam", indexed=True):
        self.pipe = pipename
        self.indexed = indexed
        self.files = safe_coerce_to_tuple(input_bams)

        self.commands = []
        self.kwargs = {"TEMP_IN_BAM": self.pipe}
        if len(self.files) > 1:
            params = picard_command(config, "MergeSamFiles")

            params.set_option("SO", "coordinate", sep="=", fixed=False)
            params.set_option("CREATE_INDEX", "False", sep="=")
            params.set_option("COMPRESSION_LEVEL", 0, sep="=")
            params.set_option("OUTPUT", "%(TEMP_OUT_BAM)s", sep="=")
            params.add_multiple_options("I", input_bams, sep="=")

            params.set_kwargs(TEMP_OUT_BAM=self.pipe)

            self.commands = [params.finalize()]
        else:
            # Ensure that the actual command depends on the input
            self.kwargs["IN_FILE_00"] = self.files[0]

            if indexed:
                self.kwargs["IN_FILE_01"] = swap_ext(self.files[0], ".bai")

    def setup(self, command):
        command.set_kwargs(**self.kwargs)


class MultiBAMInputNode(CommandNode):
    """Node which provides concatenation of input BAM files. Takes a
    MultiBAMInput object, and creates a pipe in the temporary folder which
    yields the concatenated BAM resulting from the concatenation of all input
    files. To avoid unnessary overhead, a symbolic link is used in the case
    where there is only a single input file.

    Usage example:
      class ExampleNode(MultiBAMInputNode):
        def __init__(self, config, input_bams):
            bam_input = MultiBAMInput(config, input_bams)
            command = AtomicCmd(['analyse_bam', '%(TEMP_IN_BAM)s'],
                                TEMP_IN_BAM=bam_input.pipe)
            commands = ParallelCmds(bam_input.commands + [command])
            MultiBAMInputNode.__init__(bam_input=bam_input,
                                       command=commands)
    """

    def __init__(self, bam_input, *args, **kwargs):
        self._bam_input = bam_input
        CommandNode.__init__(self, *args, **kwargs)

    def _setup(self, config, temp_root):
        CommandNode._setup(self, config, temp_root)
        dst_fname = os.path.join(temp_root, self._bam_input.pipe)
        if len(self._bam_input.files) > 1:
            os.mkfifo(dst_fname)
        else:
            src_fname, = self._bam_input.files
            os.symlink(os.path.join(os.getcwd(), src_fname), dst_fname)

            if self._bam_input.indexed:
                src_fname = os.path.join(os.getcwd(), swap_ext(src_fname, ".bai"))
                os.symlink(src_fname, dst_fname + ".bai")

    def _teardown(self, config, temp_root):
        pipe_fname = os.path.join(temp_root, self._bam_input.pipe)
        os.remove(pipe_fname)
        try_remove(pipe_fname + ".bai")
        CommandNode._teardown(self, config, temp_root)


###############################################################################

_PICARD_JAR = "picard.jar"
_PICARD_VERSION_CACHE = {}


def picard_command(config, command):
    """Returns basic AtomicJavaCmdBuilder for Picard tools commands."""
    jar_path = os.path.join(config.jar_root, _PICARD_JAR)

    if jar_path not in _PICARD_VERSION_CACHE:
        params = AtomicJavaCmdBuilder(jar_path,
                                      temp_root=config.temp_root,
                                      jre_options=config.jre_options)

        # Arbitrary command, since just '--version' does not work
        params.set_option("MarkDuplicates")
        params.set_option("--version")

        requirement = versions.Requirement(call=params.finalized_call,
                                           name="Picard tools",
                                           search=r"^(\d+)\.(\d+)",
                                           checks=versions.GE(1, 124))
        _PICARD_VERSION_CACHE[jar_path] = requirement

    version = _PICARD_VERSION_CACHE[jar_path]
    params = AtomicJavaCmdBuilder(jar_path,
                                  temp_root=config.temp_root,
                                  jre_options=config.jre_options,
                                  CHECK_JAR=version)
    params.set_option(command)

    return params



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import io
import os
import re

import pysam

from paleomix.node import \
    Node, \
    NodeError
from paleomix.common.fileutils import \
    describe_files, \
    make_dirs
from paleomix.common.utilities import \
    chain_sorted
from paleomix.common.sequences import \
    reverse_complement

import paleomix.common.formats.fastq as fastq
import paleomix.common.procs as procs
import paleomix.common.sampling as sampling
import paleomix.tools.factory as factory


class DetectInputDuplicationNode(Node):
    """Attempts to detect reads included multiple times as input based on the
    presence of reads with identical names AND sequences. This is compromise
    between sensitivity, specificity, and running time.

    A possible refinement would be to consider reads with the same name where
    one read is the prefix of the other (due to different amounts of trimming
    or collapsing of reads).
    """

    def __init__(self, input_files, output_file, dependencies=()):
        Node.__init__(self,
                      description="<Detect Input Duplication: %s>"
                      % (describe_files(input_files)),
                      input_files=input_files,
                      output_files=output_file,
                      dependencies=dependencies)

    def run(self, _):
        handles = []
        try:
            last_pos = None
            observed_reads = collections.defaultdict(list)
            for (record, filename) in self._open_samfiles(handles, self.input_files):
                curr_pos = (record.pos, record.tid)
                if curr_pos != last_pos:
                    self._process_reads(observed_reads, self.output_files)
                    observed_reads.clear()
                    last_pos = curr_pos

                observed_reads[record.qname].append((record, filename))
            self._process_reads(observed_reads, self.output_files)

            # Everything is ok, touch the output files
            for fpath in self.output_files:
                make_dirs(os.path.dirname(fpath))
                with open(fpath, "w"):
                    pass
        finally:
            for handle in handles:
                handle.close()

    @classmethod
    def _open_samfiles(cls, handles, filenames):
        sequences = []
        for filename in filenames:
            handle = pysam.Samfile(filename)
            handles.append(handle)

            sequences.append(cls._read_samfile(handle, filename))

        return chain_sorted(*sequences, key=cls._key_by_tid_pos)

    @classmethod
    def _read_samfile(cls, handle, filename):
        for record in handle:
            if record.is_unmapped and (not record.pos or record.mate_is_unmapped):
                # Ignore unmapped reads except when these are sorted
                # according to the mate position (if mapped)
                continue
            elif record.flag & 0x900:
                # Ignore supplementary / secondary alignments
                continue

            yield (record, filename)

    @classmethod
    def _process_reads(cls, observed_reads, output_files):
        for records_and_filenames in observed_reads.itervalues():
            if len(records_and_filenames) == 1:
                # Most read-names should be obseved at most once at a position
                continue

            result = collections.defaultdict(list)
            for record, filename in records_and_filenames:
                key = (record.is_reverse, record.qname, record.seq, record.qual)
                result[key].append(filename)

            for (is_reverse, name, seq, qual), filenames in result.iteritems():
                if len(filenames) == 1:
                    # Two reads had same name, but different characterstics
                    continue

                filename_counts = collections.defaultdict(int)
                for filename in filenames:
                    filename_counts[filename] += 1

                if is_reverse:
                    seq = reverse_complement(seq)
                    qual = qual[::-1]

                message = ["The same read was found multiple times!",
                           "    Name:      %r" % (name,),
                           "    Sequence:  %r" % (seq,),
                           "    Qualities: %r" % (qual,),
                           ""]

                message.append("Read was found")
                for filename, count in sorted(filename_counts.iteritems()):
                    message.append("   % 2ix in %r" % (count, filename))

                message.append("")
                message.append("This indicates that the same data files have "
                               "been included multiple times in the project. "
                               "Please review the input files used in this "
                               "project, to ensure that each set of data is "
                               "included only once!\n\n"

                               "If this is not the case, then execute the "
                               "following command(s) to mark this test as "
                               "having succeeded:")

                for fpath in output_files:
                    message.append("$ touch '%s'" % (fpath,))

                raise NodeError("\n".join(message))

    @classmethod
    def _key_by_tid_pos(cls, record):
        return (record[0].tid, record[0].pos)


class ValidateFASTQFilesNode(Node):
    def __init__(self, input_files, output_file, offset, dependencies=()):
        self._offset = offset
        Node.__init__(self,
                      description="<Validate FASTQ Files: %s>"
                      % (describe_files(input_files)),
                      input_files=input_files,
                      output_files=output_file,
                      dependencies=dependencies)

    def _run(self, _config, _temp):
        check_fastq_files(self.input_files, self._offset, True)
        output_file = tuple(self.output_files)[0]
        make_dirs(os.path.dirname(output_file))
        with open(output_file, "w"):
            pass


class ValidateFASTAFilesNode(Node):
    def __init__(self, input_files, output_file, dependencies=()):
        Node.__init__(self,
                      description="<Validate FASTA Files: %s>"
                      % (describe_files(input_files)),
                      input_files=input_files,
                      output_files=output_file,
                      dependencies=dependencies)

        assert len(self.output_files) == 1, self.output_files

    def _run(self, _config, _temp):
        for filename in self.input_files:
            check_fasta_file(filename)
        output_file, = self.output_files
        make_dirs(os.path.dirname(output_file))
        with open(output_file, "w"):
            pass


def check_fastq_files(filenames, required_offset, allow_empty=False):
    for filename in filenames:
        qualities = _read_sequences(filename)
        offsets = fastq.classify_quality_strings(qualities)
        if offsets == fastq.OFFSET_BOTH:
            raise NodeError("FASTQ file contains quality scores with both "
                            "quality offsets (33 and 64); file may be "
                            "unexpected format or corrupt. Please ensure "
                            "that this file contains valid FASTQ reads from a "
                            "single source.\n    Filename = %r" % (filename,))
        elif offsets == fastq.OFFSET_MISSING:
            if allow_empty and not qualities:
                return

            raise NodeError("FASTQ file did not contain quality scores; file "
                            "may be unexpected format or corrupt. Ensure that "
                            "the file is a FASTQ file.\n    Filename = %r"
                            % (filename,))
        elif offsets not in (fastq.OFFSET_AMBIGIOUS, required_offset):
            raise NodeError("FASTQ file contains quality scores with wrong "
                            "quality score offset (%i); expected reads with "
                            "quality score offset %i. Ensure that the "
                            "'QualityOffset' specified in the makefile "
                            "corresponds to the input.\n    Filename = %s"
                            % (offsets, required_offset, filename))


def _read_sequences(filename):
    cat_call = factory.new("cat")
    cat_call.add_multiple_values((filename,))
    cat_call = cat_call.finalized_call

    cat = None
    try:
        cat = procs.open_proc(cat_call,
                              bufsize=io.DEFAULT_BUFFER_SIZE,
                              stderr=procs.PIPE,
                              stdout=procs.PIPE)
        qualities = _collect_qualities(cat.stdout, filename)

        return sampling.reservoir_sampling(qualities, 100000)
    except:
        if cat:
            cat.kill()
            cat.wait()
            cat = None
        raise
    finally:
        rc_cat = cat.wait() if cat else 0
        if rc_cat:
            message = "Error running 'paleomix cat':\n" \
                      "  Unicat return-code = %i\n\n%s" \
                      % (rc_cat, cat.stderr.read())
            raise NodeError(message)


def _collect_qualities(handle, filename):
    header = handle.readline()
    while header:
        sequence = handle.readline()
        seperator = handle.readline()
        qualities = handle.readline()

        if not header.startswith("@"):
            if header.startswith(">"):
                raise NodeError("Input file appears to be in FASTA format "
                                "(header starts with '>', expected '@'), "
                                "but only FASTQ files are supported\n"
                                "Filename = %r" % (filename,))

            raise NodeError("Input file lacks FASTQ header (expected '@', "
                            "found %r), but only FASTQ files are supported\n"
                            "    Filename = %r" % (header[:1], filename))
        elif not qualities:
            raise NodeError("Partial record found; is not 4 lines long:\n"
                            "Filename = %r\n    Record = '%s'"
                            % (filename, header.rstrip()))
        elif not seperator.startswith("+"):
            raise NodeError("Input file lacks FASTQ seperator (expected '+', "
                            "found %r), but only FASTQ files are supported\n"
                            "    Filename = %r" % (seperator[:1], filename))
        elif len(sequence) != len(qualities):
            raise NodeError("Input file contains malformed FASTQ records; "
                            "length of sequence / qualities are not the "
                            "same.\n    Filename = %r\n    Record = '%s'"
                            % (filename, header.rstrip()))

        yield qualities
        header = handle.readline()


def check_fasta_file(filename):
    with open(filename) as handle:
        namecache = {}
        state, linelength, linelengthchanged = _NA, None, False
        for linenum, line in enumerate(handle, start=1):
            # Only \n is allowed as not all tools (e.g. GATK) handle \r
            line = line.rstrip('\n')

            if not line:
                if state in (_NA, _IN_WHITESPACE):
                    continue
                elif state == _IN_HEADER:
                    raise NodeError("Expected FASTA sequence, found empty line"
                                    "\n    Filename = %r\n    Line = %r"
                                    % (filename, linenum))
                elif state == _IN_SEQUENCE:
                    state = _IN_WHITESPACE
                else:
                    assert False
            elif line.startswith(">"):
                if state in (_NA, _IN_SEQUENCE, _IN_WHITESPACE):
                    _validate_fasta_header(filename, linenum, line, namecache)
                    state = _IN_HEADER
                    linelength = None
                    linelengthchanged = False
                elif state == _IN_HEADER:
                    raise NodeError("Empty sequences not allowed\n"
                                    "    Filename = %r\n    Line = %r"
                                    % (filename, linenum - 1))
                else:
                    assert False
            else:
                if state == _NA:
                    raise NodeError("Expected FASTA header, found %r\n"
                                    "    Filename = %r\n    Line = %r"
                                    % (line, filename, linenum))
                elif state == _IN_HEADER:
                    _validate_fasta_line(filename, linenum, line)
                    linelength = len(line)
                    state = _IN_SEQUENCE
                elif state == _IN_SEQUENCE:
                    _validate_fasta_line(filename, linenum, line)
                    # If the length has changed, then that line must be the
                    # last line in the record, which may be shorter due to the
                    # sequence length. This is because the FAI index format
                    # expects that each line has the same length.
                    if linelengthchanged or (linelength < len(line)):
                        raise NodeError("Lines in FASTQ files must be of same "
                                        "length\n    Filename = %r\n"
                                        "    Line = %r" % (filename, linenum))
                    elif linelength != len(line):
                        linelengthchanged = True
                elif state == _IN_WHITESPACE:
                    raise NodeError("Empty lines not allowed in sequences\n"
                                    "    Filename = %r\n    Line = %r"
                                    % (filename, linenum))
                else:
                    assert False

        if state in (_NA, _IN_HEADER):
            raise NodeError("File does not contain any sequences"
                            "    Filename = %r" % (filename, ))

# Standard nucleotides + UIPAC codes
_VALID_CHARS_STR = "ACGTN" "RYSWKMBDHV"
_VALID_CHARS = frozenset(_VALID_CHARS_STR.upper() + _VALID_CHARS_STR.lower())
_NA, _IN_HEADER, _IN_SEQUENCE, _IN_WHITESPACE = range(4)


def _validate_fasta_header(filename, linenum, line, cache):
    name = line.split(" ", 1)[0][1:]
    if not name:
        raise NodeError("FASTA sequence must have non-empty name\n"
                        "    Filename = %r\n    Line = %r\n"
                        % (filename, linenum))
    elif not _RE_REF_NAME.match(name):
        raise NodeError("Invalid name for FASTA sequence: %r\n"
                        "    Filename = %r\n    Line = %r\n"
                        % (name, filename, linenum))
    elif name in cache:
        raise NodeError("FASTA sequences have identical name\n"
                        "    Filename = %r\n    Name = %r\n"
                        "    Line 1 = %r\n    Line 2 = %r\n"
                        % (filename, name, linenum, cache[name]))
    cache[name] = linenum
_RE_REF_NAME = re.compile("[!-()+-<>-~][!-~]*")


def _validate_fasta_line(filename, linenum, line):
    invalid_chars = frozenset(line) - _VALID_CHARS
    if invalid_chars:
        if invalid_chars == frozenset('\r'):
            raise NodeError("FASTA file contains carriage-returns ('\\r')!\n"
                            "Please convert file to unix format, using e.g. "
                            "dos2unix.\n    Filename = %r\n" % (filename,))

        raise NodeError("FASTA sequence contains invalid characters\n"
                        "    Filename = %r\n    Line = %r\n"
                        "    Invalid characters = %r"
                        % (filename, linenum, "".join(invalid_chars)))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.nodes.picard as picard

from paleomix.common.fileutils import \
    swap_ext
from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder
from paleomix.atomiccmd.sets import \
    ParallelCmds

from paleomix.nodes.picard import \
    PicardNode, \
    ValidateBAMNode
from paleomix.nodes.samtools import \
    BAMIndexNode


def index_and_validate_bam(config, prefix, node, log_file=None,
                           create_index=True):
    input_file, has_index = _get_input_file(node)
    if not has_index and create_index:
        node = BAMIndexNode(infile=input_file,
                            dependencies=node)

    validation_params = ValidateBAMNode.customize(config=config,
                                                  input_bam=input_file,
                                                  output_log=log_file,
                                                  dependencies=node)

    # Ensure that the validation node is re-run if the index changes
    if has_index or create_index:
        bai_filename = swap_ext(input_file, ".bai")
        validation_params.command.set_kwargs(IN_BAI=bai_filename)

    # Check MD tags against reference sequence
    # FIXME: Disabled due to issues with Picard/Samtools disagreeing,
    #   backwards compatibility. See the discussion at
    #     http://sourceforge.net/mailarchive/message.php?msg_id=31348639
    # validation_params.command.set_kwargs(IN_REF=prefix["Reference"])
    # validation_params.command.add_option("R", "%(IN_REF)s", sep="=")

    # Ignored since we may filter out misses and low-quality hits during
    # mapping, which leads to a large proportion of missing PE mates.
    validation_params.command.add_option("IGNORE", "MATE_NOT_FOUND",
                                         sep="=")
    # Ignored due to high rate of false positives for lanes with few hits,
    # where high-quality reads may cause mis-identification of qualities
    validation_params.command.add_option("IGNORE",
                                         "INVALID_QUALITY_FORMAT", sep="=")

    return validation_params.build_node()


def _get_input_file(node):
    input_filename, has_index = None, False
    for filename in node.output_files:
        if filename.lower().endswith(".bai"):
            has_index = True
        elif filename.lower().endswith(".bam"):
            input_filename = filename

    return input_filename, has_index


class CleanupBAMNode(PicardNode):
    def __init__(self, config, reference, input_bam, output_bam, tags,
                 min_mapq=0, filter_unmapped=False, dependencies=()):
        flt_params = AtomicCmdBuilder(("samtools", "view", "-bu"),
                                      IN_BAM=input_bam,
                                      OUT_STDOUT=AtomicCmd.PIPE)

        if min_mapq:
            flt_params.set_option("-q", min_mapq, sep="")
        if filter_unmapped:
            flt_params.set_option("-F", "0x4", sep="")

        flt_params.add_value("%(IN_BAM)s")

        jar_params = picard.picard_command(config, "AddOrReplaceReadGroups")
        jar_params.set_option("INPUT", "/dev/stdin", sep="=")
        # Output is written to a named pipe, since the JVM may, in some cases,
        # emit warning messages to stdout, resulting in a malformed BAM.
        jar_params.set_option("OUTPUT", "%(TEMP_OUT_BAM)s", sep="=")
        jar_params.set_option("COMPRESSION_LEVEL", "0", sep="=")
        # Ensure that the BAM is sorted; this is required by the pipeline, and
        # needs to be done before calling calmd (avoiding pathologic runtimes).
        jar_params.set_option("SORT_ORDER", "coordinate", sep="=")

        # All tags are overwritten; ID is set since the default (e.g. '1')
        # causes problems with pysam due to type inference (is read as a length
        # 1 string, but written as a character).
        for tag in ("ID", "SM", "LB", "PU", "PL"):
            jar_params.set_option(tag, tags[tag], sep="=")

        jar_params.set_kwargs(IN_STDIN=flt_params,
                              TEMP_OUT_BAM="bam.pipe")

        calmd = AtomicCmdBuilder(["samtools", "calmd", "-b",
                                 "%(TEMP_IN_BAM)s", "%(IN_REF)s"],
                                 IN_REF=reference,
                                 TEMP_IN_BAM="bam.pipe",
                                 OUT_STDOUT=output_bam)

        commands = [cmd.finalize() for cmd in (flt_params, jar_params, calmd)]
        description = "<Cleanup BAM: %s -> '%s'>" \
            % (input_bam, output_bam)
        PicardNode.__init__(self,
                            command=ParallelCmds(commands),
                            description=description,
                            dependencies=dependencies)

    def _setup(self, config, temp_root):
        PicardNode._setup(self, config, temp_root)
        os.mkfifo(os.path.join(temp_root, "bam.pipe"))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import with_statement

import os
import copy
import itertools
import collections

import pysam

import paleomix.common.fileutils as fileutils
import paleomix.common.utilities as utilities
import paleomix.common.sequences as sequtils
import paleomix.common.text as text

from paleomix.common.formats.fasta import \
    FASTA
from paleomix.common.formats.msa import \
    MSA
from paleomix.node import \
    NodeError, \
    Node
from paleomix.common.bedtools import \
    BEDRecord


class CollectSequencesNode(Node):
    def __init__(self, fasta_files, sequences, destination, dependencies=()):
        """
        fasta_files -- { taxon_name_1 : filename_1, ... }
        sequences   -- { interval_name_1, ... }
        """

        self._infiles = copy.deepcopy(fasta_files)
        self._sequences = utilities.safe_coerce_to_frozenset(sequences)
        self._destination = copy.copy(destination)
        self._outfiles = [os.path.join(destination, name + ".fasta")
                          for name in self._sequences]

        input_files = list(self._infiles.itervalues())
        for filename in self._infiles.itervalues():
            input_files.append(filename + ".fai")

        desc = "<CollectSequences: %i sequences from %i files -> '%s'>" \
               % (len(self._sequences), len(self._infiles), self._destination)
        Node.__init__(self,
                      description=desc,
                      input_files=input_files,
                      output_files=self._outfiles,
                      dependencies=dependencies)

    def _setup(self, _config, _temp):
        for filename in self._infiles.itervalues():
            with open(filename + ".fai") as handle:
                sequences = set()
                for line in handle:
                    sequences.add(line.split("\t", 1)[0])

                missing_sequences = list(self._sequences - sequences)
                if missing_sequences:
                    if len(missing_sequences) >= 4:
                        missing_sequences = missing_sequences[:3]
                        missing_sequences.append("...")

                    message = ("FASTA file does not contain expected "
                               "sequences:\n  File =  %r\n  "
                               "Sequences = %s\n") \
                        % (filename, ", ".join(missing_sequences))
                    raise NodeError(message)

    def _run(self, _config, temp):
        fasta_files = []
        for (name, filename) in sorted(self._infiles.iteritems()):
            fasta_files.append((name, pysam.Fastafile(filename)))

        for sequence_name in sorted(self._sequences):
            filename = os.path.join(temp, sequence_name + ".fasta")
            with open(filename, "w") as out_handle:
                for (sample, fasta_file) in fasta_files:
                    sequence = fasta_file.fetch(sequence_name)
                    fasta = FASTA(sample, sequence_name, sequence)
                    out_handle.write(str(fasta))

    def _teardown(self, _config, temp):
        for destination in sorted(self._outfiles):
            source = fileutils.reroot_path(temp, destination)
            fileutils.move_file(source, destination)


class FilterSingletonsNode(Node):
    def __init__(self, input_file, output_file, filter_by, dependencies):
        self._input_file = input_file
        self._output_file = output_file
        self._filter_by = dict(filter_by)
        for (to_filter, groups) in self._filter_by.items():
            # The taxa to be filtered is implied to be part of the group,
            # but is not needed when actually carrying out the filtering
            groups = utilities.safe_coerce_to_frozenset(groups) \
                - utilities.safe_coerce_to_frozenset(to_filter)

            if not groups:
                raise RuntimeError("Singleton filtering must involve at least "
                                   "one other taxa")
            self._filter_by[to_filter] = groups

        Node.__init__(self,
                      description="<FilterSingleton: '%s' -> '%s'>"
                      % (input_file, output_file),
                      input_files=[input_file],
                      output_files=[output_file],
                      dependencies=dependencies)

    def _run(self, _config, temp):
        alignment = MSA.from_file(self._input_file)
        for (to_filter, groups) in self._filter_by.iteritems():
            alignment = alignment.filter_singletons(to_filter, groups)

        temp_filename = fileutils.reroot_path(temp, self._output_file)
        with open(temp_filename, "w") as handle:
            alignment.to_file(handle)
        fileutils.move_file(temp_filename, self._output_file)


class ExtractReferenceNode(Node):
    def __init__(self, reference, bedfile, outfile, dependencies=()):
        self._reference = reference
        self._bedfile = bedfile
        self._outfile = outfile

        description = "<ExtractReference: '%s' -> '%s'>" \
            % (reference, outfile)
        Node.__init__(self,
                      description=description,
                      input_files=[reference, bedfile],
                      output_files=[outfile],
                      dependencies=dependencies)

    def _run(self, _config, temp):
        def _by_name(bed):
            return bed.name

        fastafile = pysam.Fastafile(self._reference)
        seqs = collections.defaultdict(list)
        with open(self._bedfile) as bedfile:
            bedrecords = text.parse_lines_by_contig(bedfile, BEDRecord)
            for (contig, beds) in sorted(bedrecords.iteritems()):
                beds.sort(key=lambda bed: (bed.contig, bed.name, bed.start))

                for (gene, gene_beds) in itertools.groupby(beds, _by_name):
                    gene_beds = tuple(gene_beds)
                    sequence = self._collect_sequence(fastafile, gene_beds)
                    seqs[(contig, gene)] = sequence

        temp_file = os.path.join(temp, "sequences.fasta")
        with open(temp_file, "w") as out_file:
            for ((_, gene), sequence) in sorted(seqs.items()):
                FASTA(gene, None, sequence).write(out_file)

        fileutils.move_file(temp_file, self._outfile)

    @classmethod
    def _collect_sequence(cls, fastafile, beds):
        sequence = []
        for bed in beds:
            fragment = fastafile.fetch(bed.contig, bed.start, bed.end)
            if len(fragment) != (bed.end - bed.start):
                cls._report_failure(bed, fragment)

            sequence.append(fragment)
        sequence = "".join(sequence)

        if any((bed.strand == "-") for bed in beds):
            assert all((bed.strand == "-") for bed in beds)
            sequence = sequtils.reverse_complement(sequence)

        return sequence

    @classmethod
    def _report_failure(cls, bed, fragment):
        message = "Failed to extract region from " \
                  "reference sequence at %s:%i-%i; got " \
                  "%i bp, but expected %i bp." \
                  % (bed.contig, bed.start, bed.end,
                     len(fragment), (bed.end - bed.start))
        raise NodeError(message)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import random
import pickle
import operator
import nose.tools
from nose.tools import \
     assert_equal, \
     assert_raises


import paleomix.common.utilities as utils


################################################################################
################################################################################
# Tests for 'safe_coerce_to_tuple'

def test_safe_coerce_to_tuple__str():
    assert_equal(utils.safe_coerce_to_tuple("foo"), ("foo",))


def test_safe_coerce_to_tuple__unicode():
    assert_equal(utils.safe_coerce_to_tuple(u"foo"), (u"foo",))


def test_safe_coerce_to_tuple__int():
    assert_equal(utils.safe_coerce_to_tuple(17), (17,))


def test_safe_coerce_to_tuple__list():
    assert_equal(utils.safe_coerce_to_tuple([1, 3, 2]), (1, 3, 2))


def test_safe_coerce_to_tuple__tuple():
    assert_equal(utils.safe_coerce_to_tuple((1, 3, 2)), (1, 3, 2))


def test_safe_coerce_to_tuple__iterable():
    assert_equal(utils.safe_coerce_to_tuple(xrange(3)), (0, 1, 2))


def test_safe_coerce_to_tuple__dict():
    assert_equal(utils.safe_coerce_to_tuple({1: 2, 3: 4}), ({1: 2, 3: 4},))


###############################################################################
###############################################################################
# Tests for 'safe_coerce_to_frozenset'

def test_safe_coerce_to_frozenset__str():
    assert_equal(utils.safe_coerce_to_frozenset("foo"), frozenset(("foo",)))


def test_safe_coerce_to_frozenset__unicode():
    assert_equal(utils.safe_coerce_to_frozenset(u"foo"), frozenset((u"foo",)))


def test_safe_coerce_to_frozenset__int():
    assert_equal(utils.safe_coerce_to_frozenset(17), frozenset((17,)))


def test_safe_coerce_to_frozenset__list():
    assert_equal(utils.safe_coerce_to_frozenset([1, 3, 2]), frozenset((1, 3, 2)))


def test_safe_coerce_to_frozenset__tuple():
    assert_equal(utils.safe_coerce_to_frozenset((1, 3, 2)), frozenset(((1, 3, 2))))


def test_safe_coerce_to_frozenset__iterable():
    assert_equal(utils.safe_coerce_to_frozenset(xrange(3)), frozenset((0, 1, 2)))


@nose.tools.raises(TypeError)
def test_safe_coerce_to_frozenset__dict():
    utils.safe_coerce_to_frozenset({1: 2, 3: 4})


###############################################################################
###############################################################################
# Tests for 'try_cast'

def test_try_cast__int_to_int():
    assert_equal(utils.try_cast(17, int), 17)


def test_try_cast__float_to_int():
    assert_equal(utils.try_cast(17.3, int), 17)


def test_try_cast__good_str_to_int():
    assert_equal(utils.try_cast("17", int), 17)


def test_try_cast__bad_str_to_int():
    assert_equal(utils.try_cast("x17", int), "x17")


def test_try_cast__list_to_int():
    assert_equal(utils.try_cast([1, 2, 3], int), [1, 2, 3])


###############################################################################
###############################################################################
# Tests for 'crc32'

def test_crc32_is_unsigned():
    # The following is known to produce an negative value for python 2.7.2
    data = "Nobody inspects the spammish repetition"
    assert utils.crc32(data) >= 0


###############################################################################
###############################################################################
# Tests for 'set_in'

def test_set_in__single_kw_in_empty_dictionary():
    value = {}
    utils.set_in(value, ["Foo"], 17)
    assert_equal(value, {"Foo": 17})


def test_set_in__two_kws_in_empty_dictionary():
    value = {}
    utils.set_in(value, ["Foo", 13], 17)
    assert_equal(value, {"Foo": {13: 17}})


def test_set_in__three_kws_in_empty_dictionary():
    value = {}
    utils.set_in(value, ["Foo", 13, (1, 2)], 17)
    assert_equal(value, {"Foo": {13: {(1, 2): 17}}})


def test_set_in__three_kws_in_partial_dictionary():
    value = {"Foo": {12: 0}}
    utils.set_in(value, ["Foo", 13, (1, 2)], 17)
    assert_equal(value, {"Foo": {12: 0, 13: {(1, 2): 17}}})

    value = {"Foo": {13: {"Bar": None}}}
    utils.set_in(value, ["Foo", 13, (1, 2)], 17)
    assert_equal(value, {"Foo": {13: {(1, 2): 17, "Bar": None}}})


def test_set_in__update_value_one_kw():
    value = {1: None}
    utils.set_in(value, [1], 3.14)
    assert_equal(value, {1: 3.14})


def test_set_in__update_value_two_kw():
    value = {1: {2: 3}}
    utils.set_in(value, [1, 2], 365)
    assert_equal(value, {1: {2: 365}})


@nose.tools.raises(ValueError)
def test_set_in__fail_on_no_kws():
    utils.set_in({}, [], 17)


@nose.tools.raises(TypeError)
def test_set_in__fail_on_invalid_sub_dictionary_first_level():
    utils.set_in(None, [1], 17)


@nose.tools.raises(TypeError)
def test_set_in__fail_on_invalid_sub_dictionary_second_level():
    utils.set_in({1: None}, [1, 2], 17)


@nose.tools.raises(TypeError)
def test_set_in__fail_on_invalid_sub_dictionary_third_level():
    utils.set_in({1: {2: None}}, [1, 2, 3], 17)


def test_set_in__iteratable_keywords():
    value = {}
    utils.set_in(value, iter(["Foo", 13, (1, 2)]), 17)
    assert_equal(value, {"Foo": {13: {(1, 2): 17}}})


###############################################################################
###############################################################################
# Tests for 'get_in'

def test_get_in__get_value_one_keyword():
    assert_equal(utils.get_in({1: 2}, [1]), 2)


def test_get_in__get_value_two_keywords():
    assert_equal(utils.get_in({1: {2: 3}}, [1, 2]), 3)


def test_get_in__get_value_three_keywords():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [1, 2, 3]), 4)


def test_get_in__get_default_one_keyword():
    assert_equal(utils.get_in({1: 2}, [2]), None)


def test_get_in__get_default_one_keyword_with_default():
    assert_equal(utils.get_in({1: 2}, [2], "other"), "other")


def test_get_in__get_default_three_keywords_fail_at_first():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [2, 2, 4]), None)


def test_get_in__get_default_three_keywords_fail_at_first_with_default():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [2, 2, 4], "other"), "other")


def test_get_in__get_default_three_keywords_fail_at_second():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [1, 3, 4]), None)


def test_get_in__get_default_three_keywords_fail_at_second_with_default():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [1, 3, 4], "other"), "other")


def test_get_in__get_default_three_keywords_fail_at_third():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [1, 2, 4]), None)


def test_get_in__get_default_three_keywords_fail_at_third_with_default():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, [1, 2, 4], "other"), "other")


def test_get_in__iterator_keywords():
    assert_equal(utils.get_in({1: {2: {3: 4}}}, iter([1, 2, 3])), 4)


###############################################################################
###############################################################################
# Tests for 'split_before'

def _do_split(lst, key):
    # Convertion to list allows the implementation to be
    # lazy, while making comparisons for asserts easier
    return list(utils.split_before(lst, key))


def test_split_before__split_empty_list():
    assert_equal(_do_split([], None), [])


def test_split_before__split_list_with_no_true_pred():
    assert_equal(_do_split(range(10), lambda x: False), [range(10)])


def test_split_before__split_list_true_pred_at_first_position():
    assert_equal(_do_split(range(4), lambda x: x % 2 == 0), [[0, 1], [2, 3]])


def test_split_before__split_list_true_pred_at_second_position():
    assert_equal(_do_split(range(4), lambda x: x % 2 == 1), [[0], [1, 2], [3]])


def test_split_before__split_consequtive_true_pred():
    assert_equal(_do_split(range(0, 5, 2), lambda x: x % 2 == 0), [[0], [2], [4]])


def test_split_before__no_hits():
    assert_equal(_do_split(range(1, 5), lambda x: x % 5 == 0), [range(1, 5)])


###############################################################################
###############################################################################
# Tests for 'is_strictly_increasing'

def test_is_strictly_increasing__increasing_sequence():
    assert utils.is_strictly_increasing(range(100))


def test_is_strictly_increasing__non_increasing_sequence():
    lst = range(100)
    first, second = random.sample(lst, 2)
    lst[first], lst[second] = lst[second], lst[first]

    assert not utils.is_strictly_increasing(lst)


###############################################################################
###############################################################################
# Tests for 'grouper'

def test_grouper__empty_list():
    result = utils.grouper(3, [])
    assert_equal(list(result), [])


def test_grouper__non_empty_list():
    result = utils.grouper(3, range(6))
    expected = [(0, 1, 2), (3, 4, 5)]
    assert_equal(list(result), expected)


def test_grouper__non_empty_list_with_trailing():
    result = utils.grouper(3, range(7))
    expected = [(0, 1, 2), (3, 4, 5), (6, None, None)]
    assert_equal(list(result), expected)


def test_grouper__non_empty_list_with_trailing_fill_value():
    result = utils.grouper(3, range(7), fillvalue=r'\0')
    expected = [(0, 1, 2), (3, 4, 5), (6, r'\0', r'\0')]
    assert_equal(list(result), expected)


###############################################################################
###############################################################################
# Tests for 'group_by_pred'

def test_group_by_pred__empty_list():
    assert_equal(utils.group_by_pred(id, []), ([], []))


def test_group_by_pred__always_false():
    assert_equal(utils.group_by_pred(lambda x: False, [1, 2, 3]), ([], [1, 2, 3]))


def test_group_by_pred__always_true():
    assert_equal(utils.group_by_pred(lambda x: True, [1, 2, 3]), ([1, 2, 3], []))


def test_group_by_pred__is_even():
    assert_equal(utils.group_by_pred(lambda x: x % 2 == 0, [1, 2, 3]), ([2], [1, 3]))


def test_group_by_pred__iterable():
    assert_equal(utils.group_by_pred(lambda x: x % 2 == 0, xrange(1, 4)), ([2], [1, 3]))


###############################################################################
###############################################################################
# Tests for 'fragment'

def test_fragment__empty():
    assert_equal(list(utils.fragment(5, "")), [])
    assert_equal(list(utils.fragment(5, [])), [])


def test_fragment__partial_fragment():
    assert_equal(list(utils.fragment(3, "ab")), ["ab"])
    assert_equal(list(utils.fragment(3, ["a", "b"])), [["a", "b"]])


def test_fragment__single_fragment():
    assert_equal(list(utils.fragment(3, "abc")), ["abc"])
    assert_equal(list(utils.fragment(3, ["a", "b", "c"])), [["a", "b", "c"]])


def test_fragment__multiple_fragments():
    assert_equal(list(utils.fragment(3, "abcdef")), ["abc", "def"])
    assert_equal(list(utils.fragment(3, list("abcdef"))), [list("abc"), list("def")])


def test_fragment__multiple_fragments_partial():
    assert_equal(list(utils.fragment(3, "abcdefgh")), ["abc", "def", "gh"])
    assert_equal(list(utils.fragment(3, list("abcdefgh"))), [list("abc"), list("def"), list("gh")])


@nose.tools.raises(TypeError)
def test_fragment__iterable():
    list(utils.fragment(3, xrange(6)))


@nose.tools.raises(TypeError)
def test_fragment__set():
    list(utils.fragment(3, set(range(6))))


###############################################################################
###############################################################################
# Tests for 'cumsum'

def test_cumsum__empty():
    assert_equal(list(utils.cumsum([])), [])


def test_cumsum__integers():
    assert_equal(list(utils.cumsum(range(-4, 5))), [-4, -7, -9, -10, -10, -9, -7, -4, 0])


def test_cumsum__float():
    assert_equal(list(utils.cumsum((1.0, 2.0, 3.0))), [1.0, 3.0, 6.0])


def test_cumsum__initial():
    assert_equal(list(utils.cumsum(range(5), -10)), [-10, -9, -7, -4, 0])


###############################################################################
###############################################################################
# Tests for 'fast_pickle_test'

def test_fast_pickle_test__picklable():
    utils.fast_pickle_test(1)
    utils.fast_pickle_test({})
    utils.fast_pickle_test(test_cumsum__empty)


@nose.tools.raises(pickle.PicklingError)
def test_fast_pickle_test__unpicklable_1():
    _func = lambda: None  # pragma: no coverage
    utils.fast_pickle_test(_func)


@nose.tools.raises(pickle.PicklingError)
def test_fast_pickle_test__unpicklable_2():
    def _func():
        return None  # pragma: no coverage
    utils.fast_pickle_test(_func)


###############################################################################
###############################################################################
# fill_dict

def test_fill_dict__empty_dicts():
    result = utils.fill_dict({}, {})
    assert_equal(result, {})


def test_fill_dict__filling_empty_dict():
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    expected = {"a": 1, "b": {"c": 2, "d": 3}}
    result = utils.fill_dict({}, source)
    assert_equal(result, expected)


def test_fill_dict__filling_full_dict():
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    destination = {"a": 2, "b": {"c": 3, "d": 4}}
    expected = {"a": 2, "b": {"c": 3, "d": 4}}
    result = utils.fill_dict(destination, source)
    assert_equal(result, expected)


def test_fill_dict__destination_not_modified():
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    destination = {"b": {"d": 0}}
    utils.fill_dict(destination, source)
    assert_equal(destination, {"b": {"d": 0}})


def test_fill_dict__source_not_modified():
    expected = {"a": 1, "b": {"c": 2, "d": 3}}
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    destination = {"b": {"d": 0}}
    utils.fill_dict(destination, source)
    assert_equal(source, expected)


def test_fill_dict__destination_must_be_dict():
    assert_raises(TypeError, utils.fill_dict, [], {})


def test_fill_dict__source_must_be_dict():
    assert_raises(TypeError, utils.fill_dict, {}, [])


###############################################################################
###############################################################################
# chain_sorted

def test_chain_sorted__no_sequences():
    expected = ()
    result = tuple(utils.chain_sorted())
    assert_equal(expected, result)


def test_chain_sorted__single_sequence():
    sequence = (1, 2, 3)
    result = tuple(utils.chain_sorted(sequence))
    assert_equal(sequence, result)


def test_chain_sorted__sequential_contents():
    def _sequential_contents(seq_a, seq_b):
        expected = (1, 2, 3, 4, 5, 6)
        result = tuple(utils.chain_sorted(seq_a, seq_b))
        assert_equal(expected, result)

    sequence_a = (1, 2, 3)
    sequence_b = (4, 5, 6)
    yield _sequential_contents, sequence_a, sequence_b
    yield _sequential_contents, sequence_b, sequence_a


def test_chain_sorted__mixed_contents():
    sequence_a = (3, 4, 8)
    sequence_c = (0, 1, 6)
    sequence_b = (2, 5, 7)
    expected = (0, 1, 2, 3, 4, 5, 6, 7, 8)
    result = tuple(utils.chain_sorted(sequence_a, sequence_b, sequence_c))
    assert_equal(expected, result)


def test_chain_sorted__mixed_length_contents():
    sequence_a = (1,)
    sequence_c = (0, 2)
    sequence_b = ()
    expected = (0, 1, 2)
    result = tuple(utils.chain_sorted(sequence_a, sequence_b, sequence_c))
    assert_equal(expected, result)


def test_chain_sorted__mixed_contents__key():
    sequence_a = (-2, -3, -5)
    sequence_b = (0, -1, -4)
    expected = (0, -1, -2, -3, -4, -5)
    result = tuple(utils.chain_sorted(sequence_a, sequence_b, key=abs))
    assert_equal(expected, result)


def test_chain_sorted__identical_objects_are_preserved():
    object_a = [1]
    object_b = [1]
    assert object_a is not object_b
    expected = (object_a, object_b)
    result = tuple(utils.chain_sorted([object_a], [object_b]))
    assert_equal(expected, result)
    assert(object_a is result[0] or object_a is result[1])
    assert(object_b is result[0] or object_b is result[1])


def test_chain_sorted__stable_sort():
    object_a = [1]
    object_b = [1]
    object_c = [2]
    object_d = [2]
    seq_a = [object_a, object_c]
    seq_b = [object_b, object_d]

    expected = (object_a, object_b, object_c, object_d)
    result = tuple(utils.chain_sorted(seq_a, seq_b))
    assert_equal(expected, result)
    assert(all(a is b for (a, b) in zip(expected, result)))

    expected = (object_b, object_a, object_d, object_c)
    result = tuple(utils.chain_sorted(seq_b, seq_a))
    assert_equal(expected, result)
    assert(all(a is b for (a, b) in zip(expected, result)))


def test_chain_sorted__runs_of_values():
    object_a = [1]
    object_b = [1]
    object_c = [2]
    object_d = [2]
    seq_a = [object_a, object_b]
    seq_b = [object_c, object_d]

    expected = (object_a, object_b, object_c, object_d)
    result = tuple(utils.chain_sorted(seq_a, seq_b))
    assert_equal(expected, result)
    assert(all(a is b for (a, b) in zip(expected, result)))


def test_chain_sorted__invalid_keywords():
    assert_raises(TypeError, tuple, utils.chain_sorted((1, 2, 3), foobar=None))


###############################################################################
###############################################################################
# Immutable

def test_immutable__properties_set():
    class ImmutableCls(utils.Immutable):
        def __init__(self, value):
            utils.Immutable.__init__(self, value=value)

    obj = ImmutableCls(17)
    assert_equal(obj.value, 17)


def test_immutable__properties_immutable():
    def _test_immutable_property(key, value):
        class ImmutableCls(utils.Immutable):
            def __init__(self, value):
                utils.Immutable.__init__(self, value=value)

        obj = ImmutableCls(17)
        assert_raises(NotImplementedError, setattr, obj, key, value)

    yield _test_immutable_property, "value", 13
    yield _test_immutable_property, "new_value", "foo"


def test_immutable__properties_del():
    class ImmutableCls(utils.Immutable):
        def __init__(self, value):
            utils.Immutable.__init__(self, value=value)

    def _del_property(obj):
        del obj.value

    obj = ImmutableCls(17)
    assert_raises(NotImplementedError, _del_property, obj)


###############################################################################
###############################################################################
# TotallyOrdered

class SomethingOrdered(utils.TotallyOrdered):
    def __init__(self, value):
        self.value = value

    def __lt__(self, other):
        if not isinstance(other, SomethingOrdered):
            return NotImplemented

        return self.value < other.value


# Less than
def test_totally_ordered__lt_vs_lt():
    assert SomethingOrdered(1) < SomethingOrdered(2)


def test_totally_ordered__lt_vs_gt():
    assert not (SomethingOrdered(1) < SomethingOrdered(0))


def test_totally_ordered__lt_vs_wrong_type():
    assert_equal(SomethingOrdered(1).__lt__("Foo"), NotImplemented)


# Less than or equal
def test_totally_ordered__le_vs_le():
    assert SomethingOrdered(1) <= SomethingOrdered(1)


def test_totally_ordered__le_vs_gt():
    assert not (SomethingOrdered(1) <= SomethingOrdered(0))


def test_totally_ordered__le_vs_wrong_type():
    assert_equal(SomethingOrdered(1).__le__("Foo"), NotImplemented)


# Greater than or equal
def test_totally_ordered__ge_vs_ge():
    assert SomethingOrdered(1) >= SomethingOrdered(1)


def test_totally_ordered__ge_vs_lt():
    assert not (SomethingOrdered(0) >= SomethingOrdered(1))


def test_totally_ordered__ge_vs_wrong_type():
    assert_equal(SomethingOrdered(1).__ge__("Foo"), NotImplemented)


# Greater than
def test_totally_ordered__gt_vs_gt():
    assert SomethingOrdered(1) > SomethingOrdered(0)


def test_totally_ordered__gt_vs_eq():
    assert not (SomethingOrdered(0) > SomethingOrdered(0))


def test_totally_ordered__gt_vs_wrong_type():
    assert_equal(SomethingOrdered(1).__gt__("Foo"), NotImplemented)


# Equal to
def test_totally_ordered__eq_vs_eq():
    assert SomethingOrdered(1) == SomethingOrdered(1)


def test_totally_ordered__eq_vs_ne():
    assert not (SomethingOrdered(1) == SomethingOrdered(2))


def test_totally_ordered__eq_vs_wrong_type():
    assert_equal(SomethingOrdered(1).__eq__("Foo"), NotImplemented)


# Not equal to
def test_totally_ordered__ne_vs_ne():
    assert SomethingOrdered(1) != SomethingOrdered(2)


def test_totally_ordered__ne_vs_eq():
    assert not (SomethingOrdered(1) != SomethingOrdered(1))


def test_totally_ordered__ne_vs_wrong_type():
    assert_equal(SomethingOrdered(1).__ne__("Foo"), NotImplemented)


class SomethingBadlyOrdered(utils.TotallyOrdered):
    def __init__(self, value):
        self.value = value


def test_totally_ordered__missing_implementation():
    obj_a = SomethingBadlyOrdered(1)
    obj_b = SomethingBadlyOrdered(2)
    assert_raises(NotImplementedError, operator.gt, obj_a, obj_b)



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import optparse

from paleomix.common.formats.msa import \
     MSA


def _is_sufficently_covered(filepath, min_coverage):
    msa = MSA.from_file(filepath)
    if msa.seqlen() % 3:
        return False

    total_bases_not_covered = 0
    for fasta_record in msa:
        total_bases_not_covered += fasta_record.sequence.upper().count("N")
        total_bases_not_covered += fasta_record.sequence.count("-")

    total_bases = float(len(msa) * msa.seqlen())
    frac_covered = 1.0 - total_bases_not_covered / total_bases
    return frac_covered >= min_coverage


def main(argv):
    usage = "%prog [options] <PATH>"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--min-coverage", default=0.8, type=float,
                      help="Minimum fraction of called bases in a MSA "
                           "[%default]")

    config, args = parser.parse_args(argv)
    if not args:
        parser.print_usage()
        return 1

    for root_dir in args:
        for filename in os.listdir(root_dir):
            if filename.endswith(".afa"):
                fpath = os.path.join(root_dir, filename)
                if _is_sufficently_covered(fpath, config.min_coverage):
                    sequence_name, _ = filename.rsplit(".", 1)
                    print(sequence_name)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import re
import glob
import random

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    AtomicMPICmdBuilder, \
    use_customizable_cli_parameters, \
    create_customizable_cli_parameters

from paleomix.nodegraph import \
    FileStatusCache


EXAML_VERSION = versions.Requirement(call   = ("examl", "-version"),
                                     search = r"version (\d+)\.(\d+)\.(\d+)",
                                     checks = versions.GE(3, 0, 0))

PARSER_VERSION = versions.Requirement(call   = ("parse-examl", "-h"),
                                      search = r"version (\d+)\.(\d+)\.(\d+)",
                                      checks = versions.GE(3, 0, 0))


class ExaMLParserNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_alignment, input_partition, output_file, dependencies = ()):
        """
        Arguments:
        input_alignment  -- An alignment file in a format readable by RAxML.
        input_partition  -- A set of partitions in a format readable by RAxML.
        output_filename  -- Filename for the output binary sequence."""

        command = AtomicCmdBuilder("parse-examl", set_cwd = True)

        command.set_option("-s", "%(TEMP_OUT_ALN)s")
        command.set_option("-q", "%(TEMP_OUT_PART)s")
        # Output file will be named output.binary, and placed in the CWD
        command.set_option("-n", "output")

        # Substitution model
        command.set_option("-m", "DNA", fixed = False)


        command.set_kwargs(# Auto-delete: Symlinks
                          TEMP_OUT_PART   = os.path.basename(input_partition),
                          TEMP_OUT_ALN    = os.path.basename(input_alignment),

                          # Input files, are not used directly (see below)
                          IN_ALIGNMENT    = input_alignment,
                          IN_PARTITION    = input_partition,

                          # Final output file, are not created directly
                          OUT_BINARY      = output_file,

                          CHECK_EXAML     = PARSER_VERSION)

        return {"command" : command}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._symlinks = [os.path.abspath(parameters.input_alignment),
                          os.path.abspath(parameters.input_partition)]
        self._output_file = os.path.basename(parameters.output_file)


        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = "<ExaMLParser: '%s' -> '%s'>" \
                                 % (parameters.input_alignment, parameters.output_file),
                             dependencies = parameters.dependencies)


    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # Required to avoid the creation of files outside the temp folder
        for filename in self._symlinks:
            source      = os.path.abspath(filename)
            destination = os.path.join(temp, os.path.basename(filename))

            os.symlink(source, destination)


    def _teardown(self, config, temp):
        os.remove(os.path.join(temp, "RAxML_info.output"))

        source      = os.path.join(temp, "output.binary")
        destination = fileutils.reroot_path(temp, self._output_file)
        fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)


class ExaMLNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_binary, initial_tree, output_template, threads = 1, dependencies = ()):
        """
        Arguments:
        input_binary  -- A binary alignment file in a format readable by ExaML.
        output_template  -- A template string used to construct final filenames. Should consist
                            of a full path, including a single '%s', which is replaced with the
                            variable part of RAxML output files (e.g. 'info', 'bestTree', ...).
                            Example destination: '/disk/project/SN013420.RAxML.%s'
                            Example output:      '/disk/project/SN013420.RAxML.bestTree'"""

        # TODO: Make MPIParams!
        command = AtomicMPICmdBuilder("examl", threads = threads)

        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")

        command.set_option("-s", "%(IN_ALN)s")
        command.set_option("-t", "%(IN_TREE)s")
        command.set_option("-n", "Pypeline")

        command.set_kwargs(IN_ALN=input_binary,
                           IN_TREE=initial_tree,

                           # Final output files, are not created directly
                           OUT_INFO=output_template % "info",
                           OUT_BESTTREE=output_template % "result",
                           OUT_BOOTSTRAP=output_template % "log",

                           # Only generated by newer versions of ExaML
                           TEMP_OUT_MODELFILE=os.path.basename(output_template
                                                               % "modelFile"),

                           CHECK_EXAML=EXAML_VERSION)

        # Use the GAMMA model of NT substitution by default
        command.set_option("-m", "GAMMA", fixed = False)

        return {"command"         : command}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._dirname  = os.path.dirname(parameters.output_template)
        self._template = os.path.basename(parameters.output_template)

        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = "<ExaML (%i thread(s)): '%s' -> '%s'>" \
                                 % (parameters.threads,
                                    parameters.input_binary,
                                    parameters.output_template),
                             threads      = parameters.threads,
                             dependencies = parameters.dependencies)

    def _create_temp_dir(self, _config):
        """Called by 'run' in order to create a temporary folder.
        To allow restarting from checkpoints, we use a fixed folder
        determined by the output_template."""
        temp = os.path.join(self._dirname, self._template % ("temp",))
        fileutils.make_dirs(temp)
        return temp

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # The temp folder may contain old files:
        # Remove old pipes to prevent failure at _teardown
        for pipe_fname in glob.glob(os.path.join(temp, "pipe*")):
            fileutils.try_remove(pipe_fname)
        # ExaML refuses to overwrite old info files
        fileutils.try_remove(os.path.join(temp, "ExaML_info.Pypeline"))

        # Resume from last checkpoint, if one such was generated
        checkpoints = glob.glob(os.path.join(temp,
                                "ExaML_binaryCheckpoint.Pypeline_*"))
        if not checkpoints:
            return

        cache = FileStatusCache()
        if cache.files_up_to_date(checkpoints, self.input_files):
            checkpoints.sort(key=lambda fname: int(fname.rsplit("_", 1)[-1]))

            # FIXME: Less hacky solution to modifying AtomicCmds needed
            self._command._command.append("-R")
            self._command._command.append(checkpoints[-1])
        else:
            for fpath in checkpoints:
                fileutils.try_remove(fpath)

    def _teardown(self, config, temp):
        for filename in os.listdir(temp):
            match = re.match("ExaML_(.*).Pypeline", filename)
            if match:
                if "binaryCheckpoint" in match.groups():
                    os.remove(os.path.join(temp, filename))
                else:
                    source      = os.path.join(temp, filename)
                    destination = os.path.join(temp, self._template % match.groups())

                    fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)




class ParsimonatorNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_alignment, output_tree, dependencies = ()):
        """
        Arguments:
        input_alignment  -- An alignment file in a format readable by RAxML.
        output_tree      -- Filename for the output newick tree."""

        command = AtomicCmdBuilder("parsimonator", set_cwd = True)

        command.set_option("-s", "%(TEMP_OUT_ALN)s")
        command.set_option("-n", "output")
        # Random seed for the stepwise addition process
        command.set_option("-p", int(random.random() * 2**31 - 1), fixed = False)

        command.set_kwargs(# Auto-delete: Symlinks
                          TEMP_OUT_ALN   = os.path.basename(input_alignment),

                          # Input files, are not used directly (see below)
                          IN_ALIGNMENT    = input_alignment,

                          # Final output file, are not created directly
                          OUT_TREE       = output_tree)

        return {"command"         : command}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._symlinks = [os.path.abspath(parameters.input_alignment)]
        self._output_tree = os.path.basename(parameters.output_tree)


        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = "<Parsimonator: '%s' -> '%s'>" \
                                 % (parameters.input_alignment, parameters.output_tree),
                             dependencies = parameters.dependencies)


    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # Required to avoid the creation of files outside the temp folder
        for filename in self._symlinks:
            source      = os.path.abspath(filename)
            destination = os.path.join(temp, os.path.basename(filename))

            os.symlink(source, destination)


    def _teardown(self, config, temp):
        os.remove(os.path.join(temp, "RAxML_info.output"))

        source      = os.path.join(temp, "RAxML_parsimonyTree.output.0")
        destination = fileutils.reroot_path(temp, self._output_tree)
        fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)




#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function

import os
import sys
import bz2
import pysam

from optparse import OptionParser

import paleomix.tools.bam_pipeline.config as bam_cfg
import paleomix.tools.bam_pipeline.mkfile as bam_mkfile
import paleomix.tools.bam_pipeline.pipeline as bam_pipeline
from paleomix.common.fileutils import make_dirs, reroot_path, add_postfix
from paleomix.common.sequences import reverse_complement
from paleomix.common.utilities import set_in, get_in


# Indentation used when printing makefile
_INDENTATION = " " * 4


def samefile(fname_a, fname_b):
    if not (os.path.exists(fname_a) and os.path.exists(fname_b)):
        return False
    return os.path.samefile(fname_a, fname_b)


class ReadSink(object):
    # Cache of opened file-handles; used to ensure that lanes containing both
    # PE (potentialy with orphaned reads) and SE reads are all collected.
    _cache = {}

    def __init__(self, filename, handle):
        """See ReadSink.open"""
        self.filename = filename
        self._handle = handle

    def close(self):
        self._handle.close()

    def write_records(self, records):
        for record in records:
            seq = record.seq
            qual = record.qual
            if record.is_reverse:
                seq = reverse_complement(seq)
                qual = qual[::-1]

            assert len(qual) == len(seq), record.qname
            self._handle.write("@%s\n" % (record.qname,))
            self._handle.write("%s\n" % (seq,))
            self._handle.write("+\n")
            self._handle.write("%s\n" % (qual,))

    @classmethod
    def get_filename(cls, destination, prefix):
        return os.path.join(destination, "%s.fastq.bz2" % (prefix,))

    @classmethod
    def open(cls, prefix, filename):
        if filename not in cls._cache:
            handle = bz2.BZ2File(os.path.join(prefix, filename), "w")
            cls._cache[filename] = ReadSink(filename, handle)

        return cls._cache[filename]

    @classmethod
    def close_all_sinks(cls):
        for handle in cls._cache.itervalues():
            handle.close()
        cls._cache.clear()


class PEReadSink(ReadSink):
    def __init__(self, prefix, destination):
        ReadSink.__init__(self, self.get_filename(destination, "paired.{Pair}"), None)
        self._sink_se = ReadSink.open(prefix, self.get_filename(destination, "singleton"))
        self._sink_pe_1 = ReadSink.open(prefix, self.get_filename(destination, "paired.1"))
        self._sink_pe_2 = ReadSink.open(prefix, self.get_filename(destination, "paired.2"))

    def write_records(self, records):
        record_cache = {}
        for record in records:
            num = 0
            if record.is_read1:
                num = 1
            elif record.is_read2:
                num = 2
            set_in(record_cache, (record.qname, num), record)

        for pair in record_cache.itervalues():
            # Only write complete pairs
            if (1 in pair) and (2 in pair):
                self._sink_pe_1.write_records([pair.pop(1)])
                self._sink_pe_2.write_records([pair.pop(2)])

            # Any orphan files are written to the SE sink
            for record in pair.itervalues():
                self._sink_se.write_records([record])

    @classmethod
    def open(cls, prefix, destination):
        return PEReadSink(prefix, destination)


def convert_reads(config, destination, record, sink_cache):
    # Source name is used, to re-merge split lanes
    name = record.tags.get("PU_src")
    destination = os.path.join(destination, name)
    make_dirs(os.path.join(config.destination, destination))

    def _open_se_sink(reads_type):
        key = (name, reads_type)
        if not get_in(sink_cache, key):
            filename = ReadSink.get_filename(destination, reads_type.lower())
            set_in(sink_cache, key, ReadSink.open(config.destination, filename))
        return key

    for (reads_type, bam_files) in record.bams.iteritems():
        # Processed reads are pre-aligned BAMs which have been cleaned up
        if reads_type in ("Paired", "Processed"):
            # Record "Single" reads; these may result from orphan SE reads
            _open_se_sink("Singleton")

            key = (name, "Paired")
            if not get_in(sink_cache, key):
                set_in(sink_cache, key, PEReadSink.open(config.destination,
                                                        destination))
        else:
            key = _open_se_sink(reads_type)

        sink = get_in(sink_cache, key)
        for filename in bam_files:
            print("%sProcessing file %r" % (_INDENTATION * 4, filename))
            with pysam.Samfile(filename) as handle:
                def _keep_record(record):
                    return (record.qual >= config.min_quality) and \
                        (len(record.seq) >= config.min_length)

                sink.write_records(record for record in handle
                                   if _keep_record(record))


def parse_options(argv):
    tmpl = "%s <Prefix> <Makefile> [<Makefile>, ...]"

    parser = OptionParser(tmpl % "bam_pipeline remap")
    parser.add_option("--destination",
                      default="remapping", dest="destination",
                      help="Destination for resulting files [%default]")
    parser.add_option("--output-name-postfix",
                      default="_remapping", dest="postfix",
                      help="Postfix added to filenames/target names of "
                      "generated files [%default]")
    parser.add_option("--min-quality", default=0, type=int,
                      help="Minimum quality of hits to include in output "
                           "[%default]")
    parser.add_option("--min-length", default=0, type=int,
                      help="Minimum length of hits to include in output "
                           "[%default]")

    config, args = parser.parse_args(argv)
    if (len(args) < 2):
        parser.print_usage()
        return None, None

    config.prefix = args[0]

    return config, args[1:]


def main(argv):
    config, args = parse_options(argv)
    if config is None:
        return 1

    # Get default options for bam_pipeline
    bam_config, _ = bam_cfg.parse_config(args, "bam")
    makefiles = bam_pipeline.read_makefiles(bam_config, args)
    # Build .fai files for reference .fasta files
    bam_pipeline.index_references(bam_config, makefiles)

    for makefile in makefiles:
        mkfile_fname = makefile["Statistics"]["Filename"]
        bam_config.destination = os.path.dirname(mkfile_fname)
        tasks = bam_pipeline.build_pipeline_full(bam_config, makefile,
                                                 return_nodes=False)

        make_dirs(config.destination)
        makefile_name = add_postfix(makefile["Statistics"]["Filename"],
                                    config.postfix)
        makefile_path = reroot_path(config.destination, makefile_name)
        if samefile(makefile["Statistics"]["Filename"], makefile_path):
            sys.stderr.write("ERROR: Would overwrite source makefile at %r\n" % (makefile_path,))
            sys.stderr.write("       Please set --destination and/or --output-name-postfix\n")
            sys.stderr.write("       before continuing.\n")
            return 1

        print("Writing makefile", makefile_path)

        found_prefix = False
        for prefix in makefile["Prefixes"]:
            if prefix != config.prefix:
                print("%sSkipping %s" % (_INDENTATION, prefix))
            else:
                found_prefix = True

        if not found_prefix:
            sys.stderr.write("\nERROR:\n")
            sys.stderr.write("Could not find prefix %r in %r! Aborting ...\n"
                             % (config.prefix, mkfile_fname))
            return 1

        with open(makefile_path, "w") as makefile_handle:
            template = bam_mkfile.build_makefile(add_sample_tmpl=False)
            makefile_handle.write(template)
            makefile_handle.write("\n" * 3)

            for target in tasks:
                target_name = add_postfix(target.name, config.postfix)
                print("%sTarget: %s -> %s" % (_INDENTATION,
                                              target.name,
                                              target_name))

                makefile_handle.write('%s"%s":\n' % (_INDENTATION * 0,
                                                     target_name))
                for prefix in target.prefixes:
                    if prefix.name != config.prefix:
                        continue

                    for sample in prefix.samples:
                        print("%sSample: %s" % (_INDENTATION * 2, sample.name))

                        makefile_handle.write('%s"%s":\n' % (_INDENTATION * 1,
                                                             sample.name))

                        for library in sample.libraries:
                            print("%sLibrary: %s" % (_INDENTATION * 3,
                                                     library.name))
                            makefile_handle.write('%s"%s":\n'
                                                  % (_INDENTATION * 2,
                                                     library.name))

                            sink_cache = {}
                            destination = os.path.join(target_name,
                                                       "reads",
                                                       sample.name,
                                                       library.name)

                            for lane in library.lanes:
                                convert_reads(config, destination, lane, sink_cache)
                            ReadSink.close_all_sinks()

                            for lane_name in sorted(sink_cache):
                                makefile_handle.write('%s"%s":\n' % (_INDENTATION * 3, lane_name))
                                for (reads_type, sink) in sorted(sink_cache[lane_name].items()):
                                    makefile_handle.write('%s%s "%s"\n'
                                                          % (_INDENTATION * 4,
                                                             ("%s:" % (reads_type,)).ljust(20),
                                                             sink.filename))
                                makefile_handle.write("\n")
        print("\tDone ...")
        print()

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import signal
import paleomix.common.signals as signals

import nose
from nose.tools import assert_equal


def test_signal__sigterm_to_str():
    assert_equal(signals.to_str(signal.SIGTERM), "SIGTERM")


def test_signal__str_to_sigterm():
    assert_equal(signals.from_str("SIGTERM"), signal.SIGTERM)


@nose.tools.raises(KeyError)
def test_signal__to_str__unknown_signal():
    signals.to_str(1024)


@nose.tools.raises(KeyError)
def test_signal__from_str__unknown_signal():
    signals.from_str("SIGSMURF")


@nose.tools.raises(TypeError)
def test_signal__to_str__wrong_type():
    signals.to_str("SIGTERM")


@nose.tools.raises(TypeError)
def test_signal__from_str__wrong_type():
    signals.from_str(signal.SIGTERM)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import copy

import paleomix.tools.bam_pipeline.paths as paths

from paleomix.atomiccmd.builder import \
    apply_options

from paleomix.common.makefile import \
    MakefileError

from paleomix.nodes.bwa import \
    BWAAlgorithmNode, \
    BWABacktrack, \
    BWASampe, \
    BWASamse
from paleomix.nodes.bowtie2 import \
    Bowtie2Node

from paleomix.tools.bam_pipeline.parts import \
    Reads
from paleomix.tools.bam_pipeline.nodes import \
    CleanupBAMNode, \
    index_and_validate_bam

from paleomix.common.fileutils import \
    swap_ext


#
_TRIMMED_READS_CACHE = {}


class Lane:
    def __init__(self, config, prefix, record, name):
        self.name = name
        self.bams = {}
        self.reads = None
        self.options = copy.deepcopy(record["Options"])
        self.tags = tags = copy.deepcopy(record["Tags"])
        self.tags["PU"] = self.tags["PU_src"]
        self.tags["PG"] = self.tags["PG"].lower()
        self.folder = os.path.join(config.destination,
                                   tags["Target"],
                                   prefix["Name"],
                                   tags["SM"],
                                   tags["LB"],
                                   tags["PU_cur"])

        if record["Type"] == "BAMs":
            self._init_pre_aligned_lane(config, prefix, record)
        else:
            self._init_reads(config, record)
            self._init_unaligned_lane(config, prefix, record)

    def _init_pre_aligned_lane(self, config, prefix, record):
        if prefix["Name"] not in record["Data"]:
            return

        input_filename = record["Data"][prefix["Name"]]
        output_filename = os.path.join(self.folder, "processed.bam")

        node = CleanupBAMNode(config=config,
                              reference=prefix["Reference"],
                              input_bam=input_filename,
                              output_bam=output_filename,
                              tags=self.tags,
                              dependencies=prefix["Nodes"])

        index_required = self._is_indexing_required(prefix)
        validated_node = index_and_validate_bam(config, prefix, node,
                                                create_index=index_required)

        self.bams["Processed"] = {output_filename: validated_node}

    def _init_reads(self, config, record):
        key = tuple(self.tags[key] for key in ("Target", "SM", "LB", "PU_cur"))
        if key not in _TRIMMED_READS_CACHE:
            _TRIMMED_READS_CACHE[key] \
                = Reads(config, record, record["Options"]["QualityOffset"])
        self.reads = _TRIMMED_READS_CACHE[key]

    def _init_unaligned_lane(self, config, prefix, record):
        prefix_key = "Nodes:%s" % (self.options["Aligners"]["Program"],)

        for (key, input_filename) in self.reads.files.iteritems():
            # Common parameters between BWA / Bowtie2
            output_filename = os.path.join(self.folder,
                                           "%s.bam" % (key.lower(),))

            parameters = {
                "input_file": input_filename,
                "output_file": output_filename,
                "prefix": prefix["Path"],
                "reference": prefix["Reference"],
                "dependencies": self.reads.nodes + prefix[prefix_key],
            }

            nodes = self._build_alignment_nodes(config=config,
                                                record=record,
                                                prefix=prefix,
                                                parameters=parameters)

            self.bams[key] = {output_filename: nodes}

    def _build_alignment_nodes(self, config, record, prefix, parameters):
        if self.options["Aligners"]["Program"] == "BWA":
            algorithm = self.options["Aligners"]["BWA"]["Algorithm"].lower()
            parameters["threads"] = config.bwa_max_threads
            parameters["algorithm"] = algorithm

            if algorithm == "backtrack":
                return self._build_bwa_backtrack(config=config,
                                                 record=record,
                                                 prefix=prefix,
                                                 parameters=parameters)
            elif algorithm in ('mem', 'bwasw'):
                return self._build_bwa_algorithm(config=config,
                                                 record=record,
                                                 prefix=prefix,
                                                 parameters=parameters)
            else:
                raise NotImplementedError('BWA %r not implemented!'
                                          % (algorithm,))

        elif self.options["Aligners"]["Program"] == "Bowtie2":
            return self._build_bowtie2(config=config,
                                       record=record,
                                       prefix=prefix,
                                       parameters=parameters)
        else:
            raise NotImplementedError('Aligner %r not implemented!'
                                      % (self.options["Aligners"]["Program"],))

    def _is_indexing_required(self, prefix):
        """Returns true if indexing lane BAMs is nessesary.
        """
        # Indexes are required for all files when calculating region statistics
        return bool(prefix.get("RegionsOfInterest")) or \
            (self.options["Features"]["RealignedBAM"] and not
             # and for BAMs fed to GATK, but in this case we only use these
             # indexes if we don't generate PCR filtered or recaled BAMs
             (self.options["RescaleQualities"] or
              self.options["PCRDuplicates"]))

    def _set_rg_tags(self, command):
        command.set_option("--rg-id", self.tags["ID"])
        for tag_name in ("SM", "LB", "PU", "PL", "PG"):
            tag_value = "%s:%s" % (tag_name, self.tags[tag_name])
            command.add_option("--rg", tag_value)

    ###########################################################################
    ###########################################################################
    # Construction of mapping nodes

    def _build_bwa_backtrack(self, config, prefix, record, parameters):
        if paths.is_paired_end(parameters["input_file"]):
            return self._build_bwa_backtrack_pe(config=config,
                                                prefix=prefix,
                                                record=record,
                                                parameters=parameters)
        else:
            return self._build_bwa_backtrack_se(config=config,
                                                prefix=prefix,
                                                record=record,
                                                parameters=parameters)

    def _build_bwa_backtrack_aln(self, parameters, input_file, output_file):
        """
        """
        node = BWABacktrack.customize(input_file=input_file,
                                      output_file=output_file,
                                      threads=parameters["threads"],
                                      prefix=parameters["prefix"],
                                      reference=parameters["reference"],
                                      dependencies=parameters["dependencies"])

        if not self.options["Aligners"]["BWA"]["UseSeed"]:
            node.commands["aln"].set_option("-l", 2 ** 16 - 1)

        if self.options["QualityOffset"] in (64, "Solexa"):
            node.commands["aln"].set_option("-I")

        apply_options(node.commands["aln"], self.options["Aligners"]["BWA"])

        return node.build_node()

    def _build_bwa_backtrack_se(self, config, prefix, record, parameters):
        input_file_fq = parameters.pop("input_file")
        output_file_bam = parameters.pop("output_file")
        output_file_sai = swap_ext(output_file_bam, ".sai")

        aln_node = self._build_bwa_backtrack_aln(parameters=parameters,
                                                 input_file=input_file_fq,
                                                 output_file=output_file_sai)

        sam_node = BWASamse.customize(input_file_fq=input_file_fq,
                                      input_file_sai=output_file_sai,
                                      output_file=output_file_bam,
                                      prefix=parameters["prefix"],
                                      reference=parameters["reference"],
                                      dependencies=aln_node)

        return self._finalize_nodes(config, prefix, parameters, sam_node)

    def _build_bwa_backtrack_pe(self, config, prefix, record, parameters):
        template = parameters.pop("input_file")
        output_bam = parameters.pop("output_file")

        aln_files = []
        aln_nodes = []
        for mate in (1, 2):
            input_file = template.format(Pair=mate)
            output_sai = swap_ext(output_bam, "%i.sai" % (mate,))

            aln_node = self._build_bwa_backtrack_aln(parameters=parameters,
                                                     input_file=input_file,
                                                     output_file=output_sai)

            aln_files.append(output_sai)
            aln_nodes.append(aln_node)

        sam_node = BWASampe.customize(input_file_sai_1=aln_files[0],
                                      input_file_sai_2=aln_files[1],
                                      input_file_fq_1=template.format(Pair=1),
                                      input_file_fq_2=template.format(Pair=2),
                                      output_file=output_bam,
                                      prefix=parameters['prefix'],
                                      reference=parameters["reference"],
                                      dependencies=aln_nodes)

        return self._finalize_nodes(config, prefix, parameters, sam_node)

    def _build_bwa_algorithm(self, config, prefix, record, parameters):
        if self.options["QualityOffset"] != 33:
            raise MakefileError("Mapping with BWA using the %r algorithm "
                                "currently does not support QualityOffsets "
                                "other than 33; please convert your FASTQ "
                                "if you wish to proceed.")

        self._set_pe_input_files(parameters)
        node = BWAAlgorithmNode.customize(**parameters)

        return self._finalize_nodes(config, prefix, parameters, node)

    def _build_bowtie2(self, config, prefix, record, parameters):
        self._set_pe_input_files(parameters)
        node = Bowtie2Node.customize(threads=config.bowtie2_max_threads,
                                     **parameters)

        command = node.commands["aln"]
        if self.options["QualityOffset"] == 33:
            command.set_option("--phred33")
        else:
            command.set_option("--phred64")

        for (key, value) in self.options["Aligners"]["Bowtie2"].iteritems():
            if key.startswith("-"):
                command.set_option(key, value)

        return self._finalize_nodes(config, prefix, parameters, node)

    def _finalize_nodes(self, config, prefix, parameters, node):
        self._set_rg_tags(node.commands["convert"])

        min_quality = self.options["Aligners"]["BWA"]["MinQuality"]
        node.commands["convert"].set_option('-q', min_quality)

        if self.options["Aligners"]["BWA"]["FilterUnmappedReads"]:
            node.commands["convert"].set_option('-F', "0x4")

        index_required = self._is_indexing_required(prefix)
        validated_node = index_and_validate_bam(config=config,
                                                prefix=parameters['prefix'],
                                                node=node.build_node(),
                                                create_index=index_required)

        return validated_node

    @classmethod
    def _set_pe_input_files(self, parameters):
        template = parameters.pop("input_file")
        if paths.is_paired_end(template):
            parameters["input_file_1"] = template.format(Pair=1)
            parameters["input_file_2"] = template.format(Pair=2)
        else:
            parameters["input_file_1"] = template
            parameters["input_file_2"] = None



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is herby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Disable warning for missign docstrings
# pylint: disable=C0111
# Disable warning caused by "invalid" function names
# pylint: disable=C0103
# Disable warning caused by touching private member variables/functions
# pylint: disable=W0212
# Disable warnings caused by flexmock setups ("X is assigned to nothing")
# pylint: disable=W0106
import os
import random

from nose.tools import \
    assert_in, \
    assert_equal, \
    assert_raises
from flexmock import flexmock

from paleomix.common.testing import \
    with_temp_folder, \
    set_file_contents, \
    get_file_contents

from paleomix.atomiccmd.command import \
    AtomicCmd
from paleomix.node import \
    Node, \
    CommandNode, \
    NodeError, \
    NodeUnhandledException, \
    CmdNodeError
from paleomix.common.utilities import \
    safe_coerce_to_frozenset


def test_dir():
    return os.path.dirname(__file__)


def test_file(*args):
    return os.path.join(test_dir(), "data", *args)


def _CommandNodeWrap(**kwargs):
    return CommandNode(command=AtomicCmd("true"), **kwargs)
_NODE_TYPES = (Node, _CommandNodeWrap)


_DESCRIPTION = "My description of a node"
_IN_FILES = frozenset((test_file("empty_file_1"),
                       test_file("empty_file_2")))
_OUT_FILES = frozenset((test_file("missing_out_file_1"),
                        test_file("missing_out_file_2")))
_EXEC_FILES = frozenset(("ls", "sh"))
_AUX_FILES = frozenset((test_file("rCRS.fasta"),
                        test_file("rCRS.fasta.fai")))
_REQUIREMENTS = frozenset((id, str))
_EMPTY_FILE = test_file("empty_file_1")


def _build_cmd_mock(input_files  = _IN_FILES,
                    output_files = (),
                    executables  = (),
                    auxiliary_files = (),
                    requirements = (),
                    optional_temp_files = ()):
    return flexmock(input_files     = frozenset(input_files),
                    output_files    = frozenset(output_files),
                    executables     = frozenset(executables),
                    auxiliary_files = frozenset(auxiliary_files),
                    requirements    = frozenset(requirements),
                    expected_temp_files = frozenset(map(os.path.basename, output_files)),
                    optional_temp_files = frozenset(optional_temp_files))


###############################################################################
###############################################################################
# Node: Constructor: File sets

def test_constructor():
    def first(values):
        return random.choice(tuple(values))

    def _do_test_constructor__single_value(key, value):
        defaults = {"input_files": _EMPTY_FILE}
        defaults[key] = value
        node = Node(**defaults)
        expected = safe_coerce_to_frozenset(value)
        assert_equal(getattr(node, key), expected)

    # Single values
    yield _do_test_constructor__single_value, "input_files",  first(_IN_FILES)
    yield _do_test_constructor__single_value, "output_files", first(_OUT_FILES)
    yield _do_test_constructor__single_value, "executables",  first(_EXEC_FILES)
    yield _do_test_constructor__single_value, "auxiliary_files",  first(_AUX_FILES)

    # Single value in list
    yield _do_test_constructor__single_value, "input_files",  [first(_IN_FILES)]
    yield _do_test_constructor__single_value, "output_files", [first(_OUT_FILES)]
    yield _do_test_constructor__single_value, "executables",  [first(_EXEC_FILES)]
    yield _do_test_constructor__single_value, "auxiliary_files",  [first(_AUX_FILES)]

    # Multiple values in list
    yield _do_test_constructor__single_value, "input_files",  _IN_FILES
    yield _do_test_constructor__single_value, "output_files", _OUT_FILES
    yield _do_test_constructor__single_value, "executables",  _EXEC_FILES
    yield _do_test_constructor__single_value, "auxiliary_files",  _AUX_FILES


def test_constructor__invalid_values():
    def _do_test_constructor__invalid_values(key, value):
        assert_raises(TypeError, Node, **{key: value})

    yield _do_test_constructor__invalid_values, "input_files",      [id]
    yield _do_test_constructor__invalid_values, "output_files",     [-1]
    yield _do_test_constructor__invalid_values, "executables",      [{}]
    yield _do_test_constructor__invalid_values, "auxiliary_files",  [1.3]


###############################################################################
###############################################################################
# Node: Constructor: Requirements

def test_constructor__requirements():
    node = Node(requirements=id)
    assert_equal(node.requirements, frozenset([id]))
    node = Node(requirements=[id])
    assert_equal(node.requirements, frozenset([id]))
    node = Node(requirements=[id, str])
    assert_equal(node.requirements, frozenset([id, str]))


def test_constructor__requirements__wrong_type():
    def _do_test_constructor__requirements__wrong_type(value):
        assert_raises(TypeError, Node, requirements=value)

    yield _do_test_constructor__requirements__wrong_type, 17
    yield _do_test_constructor__requirements__wrong_type, {}
    yield _do_test_constructor__requirements__wrong_type, "867-5309"


###############################################################################
###############################################################################
# Node: Constructor: Dependencies

def test_constructor__nodes_is_none():
    my_node = Node(dependencies=None)
    assert_equal(my_node.dependencies, frozenset())


def test_constructor__single_node():
    sub_node = Node()
    my_node = Node(dependencies=sub_node)
    assert_equal(my_node.dependencies, frozenset([sub_node]))


def test_constructor__iterable():
    sub_nodes = [Node(), Node()]
    my_node = Node(dependencies=iter(sub_nodes))
    assert_equal(my_node.dependencies, frozenset(sub_nodes))


def test_constructor__not_a_node():
    assert_raises(TypeError, Node, dependencies=(1,))


###############################################################################
###############################################################################
# *Node: Description

def test_constructor__description():
    def _do_test_constructor__description(cls):
        my_node = cls(description=_DESCRIPTION)
        assert_equal(str(my_node), _DESCRIPTION)
    for cls in _NODE_TYPES:
        yield _do_test_constructor__description, cls


def test_constructor__description__default():
    def _do_test_constructor__description__default(cls):
        my_node = cls()
        assert_equal(str(my_node), repr(my_node))
    for cls in _NODE_TYPES:
        yield _do_test_constructor__description__default, cls


def test_constructor__description__non_string():
    def _do_test_constructor__description__non_string(cls, value):
        assert_raises(TypeError, cls, description=value)
    for cls in _NODE_TYPES:
        yield _do_test_constructor__description__non_string, cls, 1
        yield _do_test_constructor__description__non_string, cls, {}


###############################################################################
###############################################################################
# *Node: Constructor tests: #threads

def test_constructor__threads():
    def _do_test_constructor__threads(cls, nthreads):
        node = cls(threads=nthreads)
        assert_equal(node.threads, nthreads)
    for cls in (Node, _CommandNodeWrap):
        yield _do_test_constructor__threads, cls, 1L
        yield _do_test_constructor__threads, cls, 3


def test_constructor__threads_invalid_range():
    def _do_test_constructor__threads_invalid_range(cls, nthreads):
        assert_raises(ValueError, cls, threads=nthreads)
    for cls in (Node, _CommandNodeWrap):
        yield _do_test_constructor__threads_invalid_range, cls, -1
        yield _do_test_constructor__threads_invalid_range, cls, 0


def test_constructor__threads_invalid_type():
    def _do_test_constructor__threads_invalid_type(cls, nthreads):
        assert_raises(TypeError, cls, threads=nthreads)
    for cls in (Node, _CommandNodeWrap):
        yield _do_test_constructor__threads_invalid_type, cls, "1"
        yield _do_test_constructor__threads_invalid_type, cls, {}
        yield _do_test_constructor__threads_invalid_type, cls, 2.7


###############################################################################
###############################################################################
# Node: Run

_DUMMY_TEMP_ROOT = "/xyz/tmp"
_DUMMY_TEMP = os.path.join(_DUMMY_TEMP_ROOT, "xTMPx")


def test_run__order():
    cfg_mock = flexmock(temp_root=_DUMMY_TEMP_ROOT)
    node_mock = flexmock(Node())
    node_mock.should_receive("_create_temp_dir").with_args(cfg_mock).and_return(_DUMMY_TEMP).ordered.once
    node_mock.should_receive("_setup").with_args(cfg_mock, _DUMMY_TEMP).ordered.once
    node_mock.should_receive("_run").with_args(cfg_mock, _DUMMY_TEMP).ordered.once
    node_mock.should_receive("_teardown").with_args(cfg_mock, _DUMMY_TEMP).ordered.once
    node_mock.should_receive("_remove_temp_dir").with_args(_DUMMY_TEMP).ordered.once
    node_mock.run(cfg_mock)


def test_run__exceptions():
    cfg_mock = flexmock(temp_root=_DUMMY_TEMP_ROOT)

    def build_tests(key, exception, expectation):
        def test_function():
            node_mock = flexmock(Node())
            node_mock.should_receive('_create_temp_dir').with_args(cfg_mock) \
                .and_return(_DUMMY_TEMP).ordered.once
            node_mock.should_receive(key).and_raise(exception).ordered.once
            node_mock.should_receive('_remove_temp_dir').never

            assert_raises(expectation, node_mock.run, cfg_mock)

        return test_function

    print "foo"
    for key in ('_setup', '_run', '_teardown'):
        yield build_tests(key, TypeError("The castle AAARGH!"), NodeUnhandledException)
        yield build_tests(key, NodeError("He's a very naughty boy!"), NodeError)


def test_run__exception__create_temp_dir():
    cfg_mock = flexmock(temp_root=_DUMMY_TEMP_ROOT)
    node_mock = flexmock(Node())
    node_mock.should_receive('_create_temp_dir').with_args(cfg_mock) \
        .and_raise(OSError()).ordered.once

    assert_raises(NodeUnhandledException, node_mock.run, cfg_mock)


def test_run__exception__remove_temp_dir():
    cfg_mock = flexmock(temp_root=_DUMMY_TEMP_ROOT)
    node_mock = flexmock(Node())
    node_mock.should_receive('_create_temp_dir').with_args(cfg_mock) \
        .and_return(_DUMMY_TEMP).ordered.once
    node_mock.should_receive('_remove_temp_dir').with_args(_DUMMY_TEMP) \
        .and_raise(OSError()).ordered.once

    assert_raises(NodeUnhandledException, node_mock.run, cfg_mock)


def test_run__error_log__node_error():
    @with_temp_folder
    def _do_test_run__error_log__node_error(temp_folder, exception):
        temp = os.path.join(temp_folder, "xTMPx")
        cfg_mock = flexmock(temp_root=temp_folder)
        node_mock = flexmock(Node())
        node_mock.should_receive("_create_temp_dir").with_args(cfg_mock) \
            .and_return(temp).ordered.once
        node_mock.should_receive("_run").and_raise(exception).ordered.once

        os.mkdir(temp)
        assert_raises(NodeError, node_mock.run, cfg_mock)
        log_file = os.path.join(temp_folder, "xTMPx", "pipe.errors")
        assert os.path.exists(log_file)
        assert_in("Errors =", get_file_contents(log_file))

    yield _do_test_run__error_log__node_error, NodeError("ARGH!")
    yield _do_test_run__error_log__node_error, OSError("ARGH!")


###############################################################################
###############################################################################
# Node: _setup / _teardown

def test__setup__input_files():
    def _do_test__setup__input_files_exist(kwargs):
        Node(**kwargs)._setup(None, None)

    yield _do_test__setup__input_files_exist, {"executables": ("ls", "sh")}
    yield _do_test__setup__input_files_exist, {"input_files": _IN_FILES}
    yield _do_test__setup__input_files_exist, {"auxiliary_files": _IN_FILES}


def test__setup__input_files_missing():
    def _do_test__setup__input_files_exist(kwargs):
        assert_raises(NodeError, Node(**kwargs)._setup, None, None)

    yield _do_test__setup__input_files_exist, {"executables": ("ls", "shxxxx")}
    yield _do_test__setup__input_files_exist, {"input_files": _OUT_FILES}
    yield _do_test__setup__input_files_exist, {"auxiliary_files": _OUT_FILES}


def test__teardown__output_files():
    Node(input_files=_EMPTY_FILE,
         output_files=_IN_FILES)._teardown(None, None)


def test__teardown__output_files_missing():
    node = Node(input_files=_EMPTY_FILE,
                output_files=_OUT_FILES)
    assert_raises(NodeError, node._teardown, None, None)


###############################################################################
###############################################################################
# CommandNode: Constructor

_SIMPLE_DEPS = Node()
_SIMPLE_SUBS = Node()
_SIMPLE_CMD_MOCK = flexmock(input_files=_IN_FILES,
                            output_files=_OUT_FILES,
                            executables=_EXEC_FILES,
                            auxiliary_files=_AUX_FILES,
                            requirements=_REQUIREMENTS)
_SIMPLE_CMD_NODE = CommandNode(command=_SIMPLE_CMD_MOCK,
                               dependencies=_SIMPLE_DEPS)


def test_commandnode_constructor__input_files():
    assert_equal(_SIMPLE_CMD_NODE.input_files, _IN_FILES)


def test_commandnode_constructor__output_files():
    assert_equal(_SIMPLE_CMD_NODE.output_files, _OUT_FILES)


def test_commandnode_constructor__auxiliary_files():
    assert_equal(_SIMPLE_CMD_NODE.auxiliary_files, _AUX_FILES)


def test_commandnode_constructor__executables():
    assert_equal(_SIMPLE_CMD_NODE.executables, _EXEC_FILES)


def test_commandnode_constructor__requirements():
    assert_equal(_SIMPLE_CMD_NODE.requirements, _REQUIREMENTS)


def test_commandnode_constructor__dependencies():
    assert_equal(_SIMPLE_CMD_NODE.dependencies, frozenset([_SIMPLE_DEPS]))


def test_commandnode_constructor__dependencies__default():
    cmd_mock = CommandNode(command=_SIMPLE_CMD_MOCK)
    assert_equal(cmd_mock.dependencies, frozenset())


###############################################################################
###############################################################################
# CommandNode: run

def test_command_node__run():
    cfg_mock = flexmock(temp_root=_DUMMY_TEMP_ROOT)
    cmd_mock = _build_cmd_mock()
    node_mock = flexmock(CommandNode(cmd_mock))
    node_mock.should_receive("_create_temp_dir").with_args(cfg_mock) \
        .and_return(_DUMMY_TEMP).ordered.once
    node_mock.should_receive("_setup").with_args(cfg_mock, _DUMMY_TEMP).ordered.once
    cmd_mock.should_receive("run").with_args(_DUMMY_TEMP).ordered.once
    cmd_mock.should_receive("join").and_return([0]).ordered.once
    node_mock.should_receive("_teardown").with_args(cfg_mock, _DUMMY_TEMP).ordered.once
    node_mock.should_receive("_remove_temp_dir").with_args(_DUMMY_TEMP).ordered.once
    node_mock.run(cfg_mock)


###############################################################################
###############################################################################
# CommandNode: _setup()

def test_commandnode_setup__files_exist():
    def _do_test_commandnode_setup(kwargs):
        cmd_mock = _build_cmd_mock(**kwargs)
        node = CommandNode(cmd_mock)
        node._setup(None, None)
    yield _do_test_commandnode_setup, {"executables": ("ls", "sh")}
    yield _do_test_commandnode_setup, {"input_files": _IN_FILES}
    yield _do_test_commandnode_setup, {"auxiliary_files": _IN_FILES}


def test_commandnode_setup__files_missing():
    def _do_test_commandnode_setup(kwargs):
        cmd_mock = _build_cmd_mock(**kwargs)
        node = CommandNode(cmd_mock)
        assert_raises(NodeError, node._setup, None, None)

    yield _do_test_commandnode_setup, {"executables": ("ls", "shxxxxxxxxxxx")}
    yield _do_test_commandnode_setup, {"input_files": _OUT_FILES}
    yield _do_test_commandnode_setup, {"auxiliary_files": _OUT_FILES}


###############################################################################
###############################################################################
# CommandNode: _run()

def test_commandnode_run__call_order():
    cmd_mock = _build_cmd_mock()
    cmd_mock.should_receive("run").with_args("xTMPx").ordered.once
    cmd_mock.should_receive("join").with_args().and_return((0,)).ordered.once
    node = CommandNode(cmd_mock)
    node._run(None, "xTMPx")


def test_commandnode_run__exception_on_error():
    cmd_mock = _build_cmd_mock()
    cmd_mock.should_receive("run").ordered.once
    cmd_mock.should_receive("join").and_return((1,)).ordered.once
    node = CommandNode(cmd_mock)
    assert_raises(CmdNodeError, node._run, None, None)


###############################################################################
###############################################################################
# CommandNode: _teardown

def _setup_temp_folders(temp_folder):
    destination = os.path.join(temp_folder, "dst")
    temp_folder = os.path.join(temp_folder, "tmp")
    os.makedirs(temp_folder)
    os.makedirs(destination)
    return destination, temp_folder


# Commit is called on the command obj
@with_temp_folder
def test_commandnode_teardown__commit(temp_folder):
    cmd_mock = _build_cmd_mock()
    cmd_mock.should_receive("commit").with_args(temp_folder).once
    node = CommandNode(cmd_mock)
    node._teardown(None, temp_folder)


# Files exist in temp folder, and in destination after commit
@with_temp_folder
def test_commandnode_teardown(temp_folder):
    destination, temp_folder = _setup_temp_folders(temp_folder)

    cmd = AtomicCmd(("echo", "-n", "1 2 3"),
                    IN_DUMMY=_EMPTY_FILE,
                    OUT_STDOUT=os.path.join(destination, "foo.txt"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    node = CommandNode(cmd)
    assert os.path.exists(os.path.join(temp_folder, "foo.txt"))
    assert not os.path.exists(os.path.join(destination, "foo.txt"))
    node._teardown(None, temp_folder)
    assert not os.path.exists(os.path.join(temp_folder, "foo.txt"))
    assert os.path.exists(os.path.join(destination, "foo.txt"))


# Not all required files have been generated (atomic)
@with_temp_folder
def test_commandnode_teardown__missing_files_in_temp(temp_folder):
    destination, temp_folder = _setup_temp_folders(temp_folder)

    cmd = AtomicCmd(("echo", "-n", "1 2 3"),
                    IN_DUMMY=_EMPTY_FILE,
                    OUT_BAR=os.path.join(destination, "bar.txt"),
                    OUT_STDOUT=os.path.join(destination, "foo.txt"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    node = CommandNode(cmd)
    temp_files_before = set(os.listdir(temp_folder))
    dest_files_before = set(os.listdir(destination))

    assert_raises(CmdNodeError, node._teardown, None, temp_folder)
    assert_equal(temp_files_before, set(os.listdir(temp_folder)))
    assert_equal(dest_files_before, set(os.listdir(destination)))


# Not all specified TEMP_ files exist at _teardown (allowed)
@with_temp_folder
def test_commandnode_teardown__missing_optional_files(temp_folder):
    destination, temp_folder = _setup_temp_folders(temp_folder)

    cmd = AtomicCmd(("echo", "-n", "1 2 3"),
                    IN_DUMMY=_EMPTY_FILE,
                    TEMP_OUT_BAR="bar.txt",
                    OUT_STDOUT=os.path.join(destination, "foo.txt"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    node = CommandNode(cmd)
    node._teardown(None, temp_folder)
    assert_equal(os.listdir(temp_folder), [])
    assert_equal(os.listdir(destination), ["foo.txt"])


# Not all required files were in place after commit
@with_temp_folder
def _test_commandnode_teardown__missing_files_in_dest(temp_folder):
    destination, temp_folder = _setup_temp_folders(temp_folder)

    class _CmdMock(AtomicCmd):
        def commit(self, temp):
            AtomicCmd.commit(self, temp)
            os.remove(os.path.join(destination, "foo.txt"))

    cmd = _CmdMock(("touch", "%(OUT_FOO)s", "%(OUT_BAR)s"),
                   IN_DUMMY=_EMPTY_FILE,
                   OUT_FOO=os.path.join(destination, "foo.txt"),
                   OUT_BAR=os.path.join(destination, "bar.txt"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    node = CommandNode(cmd)
    assert_raises(NodeError, node._teardown, None, temp_folder)


# Unexpected files were found in the temporary directory
@with_temp_folder
def test_commandnode_teardown__extra_files_in_temp(temp_folder):
    destination, temp_folder = _setup_temp_folders(temp_folder)

    cmd = AtomicCmd(("echo", "-n", "1 2 3"),
                    IN_DUMMY=_EMPTY_FILE,
                    OUT_STDOUT=os.path.join(destination, "foo.txt"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    node = CommandNode(cmd)
    set_file_contents(os.path.join(temp_folder, "bar.txt"), "1 2 3")
    temp_files_before = set(os.listdir(temp_folder))
    dest_files_before = set(os.listdir(destination))

    assert_raises(CmdNodeError, node._teardown, None, temp_folder)
    assert_equal(temp_files_before, set(os.listdir(temp_folder)))
    assert_equal(dest_files_before, set(os.listdir(destination)))
